# Copyright 2024 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from collections import abc as collections_abc
import dataclasses
import functools
import json
import pickle
import sys
from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
from ml_collections import config_dict
import yaml


@dataclasses.dataclass
class _TestDataclass:
  string_attribute: str
  int_attribute: int


_TEST_FIELD = {'int': 0}
_TEST_DICT = {
    'float': 2.34,
    'string': 'tom',
    'int': 2,
    'list': [1, 2],
    'dict': {'float': -1.23, 'int': 23},
    'type': _TestDataclass,
}


def _test_function():
  pass


# Having ABCMeta as a metaclass shouldn't break yaml serialization.
class _TestClass(metaclass=abc.ABCMeta):

  def __init__(self):
    self.variable_1 = 1
    self.variable_2 = '2'


_test_object = _TestClass()


class _TestClassNoStr:
  pass


_TEST_DICT_BEST_EFFORT = dict(_TEST_DICT)
_TEST_DICT_BEST_EFFORT.update({
    'unserializable': _TestClass,
    'unserializable_no_str': _TestClassNoStr,
    'function': _test_function,
    'object': _test_object,
    'set': {1, 2, 3},
})

# This is how we expect the _TEST_DICT to look after we change the name float to
# double using the function configdict.recursive_rename
_TEST_DICT_CHANGE_FLOAT_NAME = {
    'double': 2.34,
    'string': 'tom',
    'int': 2,
    'list': [1, 2],
    'dict': {'double': -1.23, 'int': 23},
    'type': _TestDataclass,
}


_TEST_DICT_WITH_DATACLASS = dict(_TEST_DICT)
_TEST_DICT_WITH_DATACLASS.update({
    'dataclass': _TestDataclass(string_attribute='test_string', int_attribute=1)
})
_TEST_CONFIGDICT_WITH_DATACLASS = config_dict.ConfigDict(
    _TEST_DICT_WITH_DATACLASS
)


def _get_test_dict():
  test_dict = dict(_TEST_DICT)
  field = config_dict.FieldReference(_TEST_FIELD)
  test_dict['ref'] = field
  test_dict['ref2'] = field
  return test_dict


def _get_test_dict_best_effort():
  test_dict = dict(_TEST_DICT_BEST_EFFORT)
  field = config_dict.FieldReference(_TEST_FIELD)
  test_dict['ref'] = field
  test_dict['ref2'] = field
  return test_dict


def _get_test_config_dict():
  return config_dict.ConfigDict(_get_test_dict())


def _get_test_config_dict_best_effort():
  return config_dict.ConfigDict(_get_test_dict_best_effort())


_CLASS_NAME = 'config_dict_test'

_JSON_TEST_DICT = (
    '{"dict": {"float": -1.23, "int": 23},'
    ' "float": 2.34,'
    ' "int": 2,'
    ' "list": [1, 2],'
    ' "ref": {"int": 0},'
    ' "ref2": {"int": 0},'
    ' "string": "tom",'
    ' "type": "'
    f"<class '{_CLASS_NAME}._TestDataclass'>"
    '"}'
)
_JSON_TEST_CONFIGDICT_WITH_DATACLASS = (
    '{"dataclass": {"int_attribute": 1, "string_attribute": "test_string"},'
    ' "dict": {"float": -1.23, "int": 23},'
    ' "float": 2.34,'
    ' "int": 2,'
    ' "list": [1, 2],'
    ' "string": "tom",'
    ' "type": "'
    f"<class '{_CLASS_NAME}._TestDataclass'>"
    '"}'
)


_DICT_TYPE = "!!python/name:builtins.dict ''"
_UNSERIALIZABLE_MSG = f"<class '{_CLASS_NAME}._TestClassNoStr'>"

_TYPES = {
    'dict_type': _DICT_TYPE,
    'configdict_type': (
        '!!python/object:ml_collections.config_dict.config_dict.ConfigDict'
    ),
    'fieldreference_type': (
        '!!python/object:ml_collections.config_dict.config_dict.FieldReference'
    ),
    'testdataclass_type': f"!!python/name:{_CLASS_NAME}._TestDataclass ''",
}

_JSON_BEST_EFFORT_TEST_DICT = (
    '{"dict": {"float": -1.23, "int": 23},'
    ' "float": 2.34,'
    ' "function": "function _test_function",'
    ' "int": 2,'
    ' "list": [1, 2],'
    ' "object": {"variable_1": 1, "variable_2": "2"},'
    ' "ref": {"int": 0},'
    ' "ref2": {"int": 0},'
    ' "set": [1, 2, 3],'
    ' "string": "tom",'
    ' "type": "'
    f"<class '{_CLASS_NAME}._TestDataclass'>\","
    ' "unserializable": "'
    f"<class '{_CLASS_NAME}._TestClass'>\","
    f' "unserializable_no_str": "{_UNSERIALIZABLE_MSG}"}}'
)

_REPR_TEST_DICT = """
dict:
  float: -1.23
  int: 23
float: 2.34
int: 2
list:
- 1
- 2
ref: &id001 {fieldreference_type}
  _field_type: {dict_type}
  _ops: []
  _required: false
  _value:
    int: 0
ref2: *id001
string: tom
type: {testdataclass_type}
""".format(**_TYPES)

_STR_TEST_DICT = """
dict: {float: -1.23, int: 23}
float: 2.34
int: 2
list: [1, 2]
ref: &id001 {int: 0}
ref2: *id001
string: tom
"""

_STR_NESTED_TEST_DICT = """
dict: {float: -1.23, int: 23}
float: 2.34
int: 2
list: [1, 2]
nested_dict:
  float: -1.23
  int: 23
  nested_dict:
    float: -1.23
    int: 23
    non_nested_dict: {float: -1.23, int: 23}
  nested_list:
  - 1
  - 2
  - [3, 4, 5]
  - 6
ref: &id001 {int: 0}
ref2: *id001
string: tom
"""


class ConfigDictTest(parameterized.TestCase):
  """Tests ConfigDict in config flags library."""

  def assertEqualConfigs(self, cfg, dictionary):
    """Asserts recursive equality of config and a dictionary."""
    self.assertEqual(cfg.to_dict(), dictionary)

  def testCreating(self):
    """Tests basic config creation."""
    cfg = config_dict.ConfigDict()
    cfg.field = 2.34
    self.assertEqual(cfg.field, 2.34)

  def testDir(self):
    """Test that dir() works correctly on config."""
    cfg = config_dict.ConfigDict()
    cfg.field = 2.34
    self.assertIn('field', dir(cfg))
    self.assertIn('lock', dir(cfg))

  def testFromDictConstruction(self):
    """Tests creation of config from existing dictionary."""
    cfg = config_dict.ConfigDict(_TEST_DICT)
    self.assertEqualConfigs(cfg, _TEST_DICT)

  def testOverridingValues(self):
    """Tests basic values overriding."""
    cfg = config_dict.ConfigDict()
    cfg.field = 2.34
    self.assertEqual(cfg.field, 2.34)
    cfg.field = -2.34
    self.assertEqual(cfg.field, -2.34)

  def testDictAttributeTurnsIntoConfigDict(self):
    """Tests that dicts in a ConfigDict turn to ConfigDicts (recursively)."""
    cfg = config_dict.ConfigDict(_TEST_DICT)
    # Test conversion to dict on creation.
    self.assertIsInstance(cfg.dict, config_dict.ConfigDict)

    # Test conversion to dict on setting attribute.
    new_dict = {'inside_dict': {'inside_key': 0}}
    cfg.new_dict = new_dict
    self.assertIsInstance(cfg.new_dict, config_dict.ConfigDict)
    self.assertIsInstance(cfg.new_dict.inside_dict, config_dict.ConfigDict)
    self.assertEqual(cfg.new_dict.to_dict(), new_dict)

  def testOverrideExceptions(self):
    """Test the `int` and unicode-string exceptions to overriding.

    ConfigDict forces strong type-checking with two exceptions. The first is
    that `int` values can be stored to fields of type `float`. And secondly,
    all string types can be  stored in fields of type `str` or `unicode`.
    """
    cfg = config_dict.ConfigDict()
    # Test that overriding 'float' fields with int works.
    cfg.float_field = 2.34
    cfg.float_field = 2
    self.assertEqual(cfg.float_field, 2.0)
    # Test that overriding with Unicode strings works.
    cfg.string_field = '42'
    cfg.string_field = '42'
    self.assertEqual(cfg.string_field, '42')
    # Test that overriding a Unicode field with a `str` type works.
    cfg.unicode_string_field = '42'
    cfg.unicode_string_field = '42'
    self.assertEqual(cfg.unicode_string_field, '42')
    # Test that overriding a list with a tuple works.
    cfg.tuple_field = [1, 2, 3]
    cfg.tuple_field = (1, 2)
    self.assertEqual(cfg.tuple_field, [1, 2])
    # Test that overriding a tuple with a list works.
    cfg.list_field = [23, 42]
    cfg.list_field = (8, 9, 10)
    self.assertEqual(cfg.list_field, [8, 9, 10])
    # Test that int <-> long conversions work.
    int_value = 1
    # In Python 2, int(very large number) returns a long
    long_value = int(1e100)
    cfg.int_field = int_value
    cfg.int_field = long_value
    self.assertEqual(cfg.int_field, long_value)
    self.assertIsInstance(cfg.int_field, int)
    cfg.long_field = long_value
    cfg.long_field = int_value
    self.assertEqual(cfg.long_field, int_value)
    self.assertIsInstance(cfg.long_field, int)

  def testOverrideCallable(self):
    """Test that overriding a callable with a callable works."""

    class SomeClass:

      def __init__(self, x, power=1):
        self.y = x**power

      def factory(self, x):
        return SomeClass(self.y + x)

    fn1 = SomeClass
    fn2 = lambda x: SomeClass(x, power=2)
    fn3 = functools.partial(SomeClass, power=3)
    fn4 = SomeClass(4.0).factory
    cfg = config_dict.ConfigDict()
    for orig in [fn1, fn2, fn3, fn4]:
      for new in [fn1, fn2, fn3, fn4]:
        cfg.fn_field = orig
        cfg.fn_field = new
        self.assertEqual(cfg.fn_field, new)

  def testOverrideFieldReference(self):
    """Test overriding with FieldReference objects."""
    cfg = config_dict.ConfigDict()
    cfg.field_1 = 'field_1'
    cfg.field_2 = 'field_2'
    # Override using a FieldReference.
    cfg.field_1 = config_dict.FieldReference('override_1')
    # Override FieldReference field using another FieldReference.
    cfg.field_1 = config_dict.FieldReference('override_2')
    # Override using empty FieldReference.
    cfg.field_2 = config_dict.FieldReference(None, field_type=str)
    # Override FieldReference field using string.
    cfg.field_2 = 'field_2'

    # Check a TypeError is raised when using FieldReference's with wrong type.
    with self.assertRaises(TypeError):
      cfg.field_2 = config_dict.FieldReference(1)

    with self.assertRaises(TypeError):
      cfg.field_2 = config_dict.FieldReference(None, field_type=int)

  def testTypeSafe(self):
    """Tests type safe checking."""
    cfg = _get_test_config_dict()

    with self.assertRaisesRegex(TypeError, "field 'float'"):
      cfg.float = 'tom'

    # Test that float cannot be assigned to int.
    with self.assertRaisesRegex(TypeError, "field 'int'"):
      cfg.int = 12.8

    with self.assertRaisesRegex(TypeError, "field 'string'"):
      cfg.string = -123

    with self.assertRaisesRegex(TypeError, "field 'float'"):
      cfg.dict.float = 'string'

    # Ensure None is ignored by type safety
    cfg.string = None
    cfg.string = 'tom'

  def testIgnoreType(self):
    cfg = config_dict.ConfigDict({
        'string': 'This is a string',
        'float': 3.0,
        'list': [config_dict.ConfigDict({'float': 1.0})],
        'tuple': [config_dict.ConfigDict({'float': 1.0})],
        'dict': {'float': 1.0},
    })

    with cfg.ignore_type():
      cfg.string = -123
      cfg.float = 'string'
      cfg.list[0].float = 'string'
      cfg.tuple[0].float = 'string'
      cfg.dict.float = 'string'

  def testTypeUnsafe(self):
    """Tests lack of type safe checking."""
    cfg = config_dict.ConfigDict(_get_test_dict(), type_safe=False)
    cfg.float = 'tom'
    cfg.string = -123
    cfg.int = 12.8

  def testLocking(self):
    """Tests lock mechanism."""
    cfg = config_dict.ConfigDict()
    cfg.field = 2
    cfg.dict_field = {'float': 1.23, 'integer': 3}
    cfg.ref = config_dict.FieldReference(config_dict.ConfigDict({'integer': 0}))
    cfg.lock()

    cfg.field = -4

    with self.assertRaises(AttributeError):
      cfg.new_field = 2

    with self.assertRaises(AttributeError):
      cfg.dict_field.new_field = -1.23

    with self.assertRaises(AttributeError):
      cfg.ref.new_field = 1

    with self.assertRaises(AttributeError):
      del cfg.field

  def testUnlocking(self):
    """Tests unlock mechanism."""
    cfg = config_dict.ConfigDict()
    cfg.dict_field = {'float': 1.23, 'integer': 3}
    cfg.ref = config_dict.FieldReference(config_dict.ConfigDict({'integer': 0}))
    cfg.lock()
    with cfg.unlocked():
      cfg.new_field = 2
      cfg.dict_field.new_field = -1.23
      cfg.ref.new_field = 1

  def testGetMethod(self):
    """Tests get method."""
    cfg = _get_test_config_dict()
    self.assertEqual(cfg.get('float', -1), cfg.float)
    self.assertEqual(cfg.get('ref', -1), cfg.ref)
    self.assertEqual(cfg.get('another_key', -1), -1)
    self.assertIsNone(cfg.get('another_key'))

  def testItemsMethod(self):
    """Tests items method."""
    cfg = _get_test_config_dict()
    self.assertEqual(dict(**cfg), dict(cfg.items()))
    items = cfg.items()
    self.assertEqual(len(items), len(_get_test_dict()))
    for entry in _TEST_DICT.items():
      if isinstance(entry[1], dict):
        entry = (entry[0], config_dict.ConfigDict(entry[1]))
      self.assertIn(entry, items)
    self.assertIn(('ref', cfg.ref), items)
    self.assertIn(('ref2', cfg.ref2), items)
    ind_ref = items.index(('ref', cfg.ref))
    ind_ref2 = items.index(('ref2', cfg.ref2))
    self.assertIs(items[ind_ref][1], items[ind_ref2][1])
    cfg = config_dict.ConfigDict()
    self.assertEqual(dict(**cfg), dict(cfg.items()))
    # Test that items are sorted
    self.assertEqual(sorted(dict(**cfg).items()), cfg.items())

  def testGetItemRecursively(self):
    """Tests getting items recursively (e.g., config['a.b'])."""
    cfg = _get_test_config_dict()
    self.assertEqual(cfg['dict.float'], -1.23)
    self.assertEqual('%(dict.int)i' % cfg, '23')

  def testIterItemsMethod(self):
    """Tests iteritems method."""
    cfg = _get_test_config_dict()
    self.assertEqual(dict(**cfg), dict(cfg.iteritems()))
    cfg = config_dict.ConfigDict()
    self.assertEqual(dict(**cfg), dict(cfg.iteritems()))

  def testIterKeysMethod(self):
    """Tests iterkeys method."""
    some_dict = {'x1': 32, 'x2': 5.2, 'x3': 'str'}
    cfg = config_dict.ConfigDict(some_dict)
    self.assertEqual(set(some_dict.keys()), set(cfg.keys()))
    # Test that keys are sorted
    for k_ref, k in zip(sorted(cfg.keys()), cfg.keys()):
      self.assertEqual(k_ref, k)

  def testKeysMethod(self):
    """Tests keys method."""
    some_dict = {'x1': 32, 'x2': 5.2, 'x3': 'str'}
    cfg = config_dict.ConfigDict(some_dict)
    self.assertEqual(set(some_dict.keys()), set(cfg.keys()))
    # Test that keys are sorted
    for k_ref, k in zip(sorted(cfg.keys()), cfg.keys()):
      self.assertEqual(k_ref, k)

  def testLenMethod(self):
    """Tests keys method."""
    some_dict = {'x1': 32, 'x2': 5.2, 'x3': 'str'}
    cfg = config_dict.ConfigDict(some_dict)
    self.assertLen(cfg, len(some_dict))

  def testIterValuesMethod(self):
    """Tests itervalues method."""
    some_dict = {'x1': 32, 'x2': 5.2, 'x3': 'str'}
    cfg = config_dict.ConfigDict(some_dict)
    self.assertEqual(set(some_dict.values()), set(cfg.values()))
    # Test that items are sorted
    for k_ref, v in zip(sorted(cfg.keys()), cfg.values()):
      self.assertEqual(cfg[k_ref], v)

  def testValuesMethod(self):
    """Tests values method."""
    some_dict = {'x1': 32, 'x2': 5.2, 'x3': 'str'}
    cfg = config_dict.ConfigDict(some_dict)
    self.assertEqual(set(some_dict.values()), set(cfg.values()))
    # Test that items are sorted
    for k_ref, v in zip(sorted(cfg.keys()), cfg.values()):
      self.assertEqual(cfg[k_ref], v)

  def testIterValuesResolvesReferences(self):
    """Tests itervalues FieldReference resolution."""
    cfg = config_dict.ConfigDict({'x1': 32, 'x2': 5.2, 'x3': 'str'})
    ref = config_dict.FieldReference(0)
    cfg['x4'] = ref
    for v in cfg.itervalues():
      self.assertNotIsInstance(v, config_dict.FieldReference)
    self.assertIn(ref, cfg.itervalues(preserve_field_references=True))

  def testValuesResolvesReferences(self):
    """Tests values FieldReference resolution."""
    cfg = config_dict.ConfigDict({'x1': 32, 'x2': 5.2, 'x3': 'str'})
    ref = config_dict.FieldReference(0)
    cfg['x4'] = ref
    for v in cfg.values():
      self.assertNotIsInstance(v, config_dict.FieldReference)
    self.assertIn(ref, cfg.values(preserve_field_references=True))

  def testIterItemsResolvesReferences(self):
    """Tests iteritems FieldReference resolution."""
    cfg = config_dict.ConfigDict({'x1': 32, 'x2': 5.2, 'x3': 'str'})
    ref = config_dict.FieldReference(0)
    cfg['x4'] = ref
    for _, v in cfg.iteritems():
      self.assertNotIsInstance(v, config_dict.FieldReference)
    self.assertIn(('x4', ref), cfg.iteritems(preserve_field_references=True))

  def testItemsResolvesReferences(self):
    """Tests items FieldReference resolution."""
    cfg = config_dict.ConfigDict({'x1': 32, 'x2': 5.2, 'x3': 'str'})
    ref = config_dict.FieldReference(0)
    cfg['x4'] = ref
    for _, v in cfg.items():
      self.assertNotIsInstance(v, config_dict.FieldReference)
    self.assertIn(('x4', ref), cfg.items(preserve_field_references=True))

  def testIntKeys(self):
    """Tests items FieldReference resolution."""
    data = {
        0: 1,
        1: 2,
        2: 3,
        'a': 4,
    }
    cfg = config_dict.ConfigDict(data, sort_keys=False)
    self.assertEqual(cfg.to_dict(), data)

  def testEquals(self):
    """Tests __eq__ and __ne__ methods."""
    some_dict = {
        'float': 1.23,
        'integer': 3,
        'list': [1, 2],
        'dict': {'a': {}, 'b': 'string'},
    }
    cfg = config_dict.ConfigDict(some_dict)
    cfg_other = config_dict.ConfigDict(some_dict)
    self.assertEqual(cfg, cfg_other)
    self.assertEqual(config_dict.ConfigDict(cfg), cfg_other)
    cfg_other.float = 3
    self.assertNotEqual(cfg, cfg_other)
    cfg_other.float = cfg.float
    cfg_other.list = ['a', 'b']
    self.assertNotEqual(cfg, cfg_other)
    cfg_other.list = cfg.list
    cfg_other.lock()
    self.assertNotEqual(cfg, cfg_other)
    cfg_other.unlock()
    cfg_other = config_dict.ConfigDict(some_dict, type_safe=False)
    self.assertNotEqual(cfg, cfg_other)
    cfg = config_dict.ConfigDict(some_dict)

    # References that have the same id should be equal (even if self-references)
    cfg_other = config_dict.ConfigDict(some_dict)
    cfg_other.me = cfg
    cfg.me = cfg
    self.assertEqual(cfg, cfg_other)
    cfg = config_dict.ConfigDict(some_dict)
    cfg.me = cfg
    self.assertEqual(cfg, cfg)

    # Self-references that do not have the same id loop infinitely
    cfg_other = config_dict.ConfigDict(some_dict)
    cfg_other.me = cfg_other

    # Temporarily disable coverage trace while testing runtime is exceeded
    trace_func = sys.gettrace()
    sys.settrace(None)

    with self.assertRaises(RuntimeError):
      cfg == cfg_other  # pylint:disable=pointless-statement

    sys.settrace(trace_func)

  def testEqAsConfigDict(self):
    """Tests .eq_as_configdict() method."""
    cfg_1 = _get_test_config_dict()
    cfg_2 = _get_test_config_dict()
    cfg_2.added_field = 3.14159
    cfg_self_ref = _get_test_config_dict()
    cfg_self_ref.self_ref = cfg_self_ref
    frozen_cfg_1 = config_dict.FrozenConfigDict(cfg_1)
    frozen_cfg_2 = config_dict.FrozenConfigDict(cfg_2)

    self.assertTrue(cfg_1.eq_as_configdict(cfg_1))
    self.assertTrue(cfg_1.eq_as_configdict(frozen_cfg_1))
    self.assertTrue(frozen_cfg_1.eq_as_configdict(cfg_1))
    self.assertTrue(frozen_cfg_1.eq_as_configdict(frozen_cfg_1))

    self.assertFalse(cfg_1.eq_as_configdict(cfg_2))
    self.assertFalse(cfg_1.eq_as_configdict(frozen_cfg_2))
    self.assertFalse(frozen_cfg_1.eq_as_configdict(cfg_self_ref))
    self.assertFalse(frozen_cfg_1.eq_as_configdict(frozen_cfg_2))
    self.assertFalse(cfg_self_ref.eq_as_configdict(cfg_1))

  def testHash(self):
    some_dict = {'float': 1.23, 'integer': 3}
    cfg = config_dict.ConfigDict(some_dict)
    with self.assertRaisesRegex(TypeError, 'unhashable type'):
      hash(cfg)

    # Ensure Python realizes ConfigDict is not hashable.
    self.assertNotIsInstance(cfg, collections_abc.Hashable)

  def testDidYouMeanFeature(self):
    """Tests 'did you mean' suggestions."""
    cfg = config_dict.ConfigDict()
    cfg.learning_rate = 0.01
    cfg.lock()

    with self.assertRaisesRegex(
        AttributeError, 'Did you mean.*learning_rate.*'
    ):
      _ = cfg.laerning_rate

    with cfg.unlocked():
      with self.assertRaisesRegex(
          AttributeError, 'Did you mean.*learning_rate.*'
      ):
        del cfg.laerning_rate

    with self.assertRaisesRegex(
        AttributeError, 'Did you mean.*learning_rate.*'
    ):
      cfg.laerning_rate = 0.02

    self.assertEqual(cfg.learning_rate, 0.01)
    with self.assertRaises(AttributeError):
      _ = self.laerning_rate

  def testReferences(self):
    """Tests assigning references in the dict."""
    cfg = _get_test_config_dict()
    cfg.dict_ref = cfg.dict

    self.assertEqual(cfg.dict_ref, cfg.dict)

  def testPreserveReferences(self):
    """Tests that initializing with another ConfigDict preserves references."""
    cfg = _get_test_config_dict()
    # In the original, "ref" and "ref2" are the same FieldReference
    self.assertIs(cfg.get_ref('ref'), cfg.get_ref('ref2'))

    # Create a copy from the original
    cfg2 = config_dict.ConfigDict(cfg)
    # If the refs had not been preserved, get_ref would create a new
    # reference for each call
    self.assertIs(cfg2.get_ref('ref'), cfg2.get_ref('ref2'))
    self.assertIs(cfg2.ref, cfg2.ref2)  # the values are also the same object

  def testUnpacking(self):
    """Tests ability to pass ConfigDict instance with ** operator."""
    cfg = config_dict.ConfigDict()
    cfg.x = 2

    def foo(x):
      return x + 3

    self.assertEqual(foo(**cfg), 5)

  def testUnpackingWithFieldReference(self):
    """Tests ability to pass ConfigDict instance with ** operator."""
    cfg = config_dict.ConfigDict()
    cfg.x = config_dict.FieldReference(2)

    def foo(x):
      return x + 3

    self.assertEqual(foo(**cfg), 5)

  def testReadingIncorrectField(self):
    """Tests whether accessing non-existing fields raises an exception."""
    cfg = config_dict.ConfigDict()
    with self.assertRaises(AttributeError):
      _ = cfg.non_existing_field

    with self.assertRaises(KeyError):
      _ = cfg['non_existing_field']

  def testIteration(self):
    """Tests whether one can iterate over ConfigDict."""
    cfg = config_dict.ConfigDict()
    for i in range(10):
      cfg['field{}'.format(i)] = 'field{}'.format(i)

    for field in cfg:
      self.assertEqual(cfg[field], getattr(cfg, field))

  def testDeYaml(self):
    """Tests YAML deserialization."""
    cfg = _get_test_config_dict()
    deyamled = yaml.load(cfg.to_yaml(), yaml.UnsafeLoader)
    self.assertEqual(cfg, deyamled)

  def testJSONConversion(self):
    """Tests JSON serialization."""
    cfg = _get_test_config_dict()
    self.assertEqual(
        cfg.to_json(sort_keys=True).strip(), _JSON_TEST_DICT.strip()
    )

    cfg = _get_test_config_dict_best_effort()
    with self.assertRaises(TypeError):
      cfg.to_json()

  def testJSONConversionCustomEncoder(self):
    """Tests JSON serialization with custom encoder."""
    cfg = _get_test_config_dict()
    encoder = json.JSONEncoder()
    mock_encoder_cls = mock.MagicMock()
    mock_encoder_cls.return_value = encoder
    with mock.patch.object(encoder, 'default') as mock_default:
      mock_default.return_value = ''
      cfg.to_json(json_encoder_cls=mock_encoder_cls)
      mock_default.assert_called()

  def testJSONConversionBestEffort(self):
    """Tests JSON serialization."""
    # Check that best effort option doesn't break default functionality
    cfg = _get_test_config_dict()
    self.assertEqual(
        cfg.to_json_best_effort(sort_keys=True).strip(), _JSON_TEST_DICT.strip()
    )

    cfg_best_effort = _get_test_config_dict_best_effort()
    self.assertEqual(
        cfg_best_effort.to_json_best_effort(sort_keys=True).strip(),
        _JSON_BEST_EFFORT_TEST_DICT.strip(),
    )

  def testJSONConversionWithDataclass(self):
    """Tests JSON serialization when the configdict contains a dataclass."""
    self.assertEqual(
        _TEST_CONFIGDICT_WITH_DATACLASS.to_json_best_effort(
            sort_keys=True
        ).strip(),
        _JSON_TEST_CONFIGDICT_WITH_DATACLASS.strip(),
    )

  def testJSONConversionTopLevel(self):
    """Tests JSON serialization."""

    class JSONEncoder(config_dict.CustomJSONEncoder):
      """JSON encoder for ConfigDict and FieldReference.

      The encoder throws an exception for non-supported types.
      """

      def default(self, obj):
        if isinstance(obj, config_dict.ConfigDict):
          return {'__serialized__': True, **obj._fields}
        return super().default(obj)

    cfg = config_dict.ConfigDict({
        'a': {'c': 123},
        'b': 456,
    })
    # Both the top-level dict and inner dict have the additional
    # `__serialized__` field.
    self.assertEqual(
        json.loads(cfg.to_json(json_encoder_cls=JSONEncoder)),
        {
            'a': {'c': 123, '__serialized__': True},
            'b': 456,
            '__serialized__': True,
        },
    )

  def testReprConversion(self):
    """Tests repr conversion."""
    cfg = _get_test_config_dict()
    self.assertEqual(repr(cfg).strip(), _REPR_TEST_DICT.strip())

  def testLoadFromRepr(self):
    cfg_dict = config_dict.ConfigDict()
    field = config_dict.FieldReference(1)
    cfg_dict.r1 = field
    cfg_dict.r2 = field

    cfg_load = yaml.load(repr(cfg_dict), yaml.UnsafeLoader)

    # Test FieldReferences are preserved
    cfg_load['r1'].set(2)
    self.assertEqual(cfg_load['r1'].get(), cfg_load['r2'].get())

  def testStrConversion(self):
    """Tests conversion to str."""
    cfg = _get_test_config_dict()
    # Verify srt(cfg) doesn't raise errors.
    _ = str(cfg)

    test_dict_2 = _get_test_dict()
    test_dict_2['nested_dict'] = {
        'float': -1.23,
        'int': 23,
        'nested_dict': {
            'float': -1.23,
            'int': 23,
            'non_nested_dict': {
                'float': -1.23,
                'int': 233,
            },
        },
        'nested_list': [1, 2, [3, 44, 5], 6],
    }
    cfg_2 = config_dict.ConfigDict(test_dict_2)
    # Demonstrate that dot-access works.
    cfg_2.nested_dict.nested_dict.non_nested_dict.int = 23
    cfg_2.nested_dict.nested_list[2][1] = 4
    # Verify srt(cfg) doesn't raise errors.
    _ = str(cfg_2)

  def testDotInField(self):
    """Tests trying to create a dot containing field."""
    cfg = config_dict.ConfigDict()

    with self.assertRaises(ValueError):
      cfg['invalid.name'] = 2.3

  def testDotInFieldAllowedNested(self):
    """Tests trying to create a dot containing field."""
    sub_dict = {'x': 0, 'y.z': 7}
    some_dict = {'a': 3, 'subcfg': sub_dict, 'b.c': 2}
    cfg = config_dict.ConfigDict(some_dict, allow_dotted_keys=True)
    self.assertTrue(cfg.allow_dotted_keys)
    self.assertTrue(cfg.subcfg.allow_dotted_keys)
    self.assertEqual(cfg.a, 3)
    self.assertEqual(cfg.subcfg.x, 0)
    self.assertEqual(cfg.subcfg['y.z'], 7)
    self.assertEqual(cfg['b.c'], 2)

    with self.assertRaises(AttributeError):
      cfg.subcfg.y.z = 5

    with self.assertRaises(AttributeError):
      cfg.b.c = 6

  def testToDictConversion(self):
    """Tests whether references are correctly handled when calling to_dict."""
    cfg = config_dict.ConfigDict()

    field = config_dict.FieldReference('a string')
    cfg.dict = {
        'float': 2.3,
        'integer': 1,
        'field_ref1': field,
        'field_ref2': field,
    }
    cfg.ref = cfg.dict
    cfg.self_ref = cfg

    pure_dict = cfg.to_dict()
    self.assertEqual(type(pure_dict), dict)

    self.assertIs(pure_dict, pure_dict['self_ref'])
    self.assertIs(pure_dict['dict'], pure_dict['ref'])

    # Ensure ConfigDict has been converted to dict.
    self.assertEqual(type(pure_dict['dict']), dict)

    # Ensure FieldReferences are not preserved, by default.
    self.assertNotIsInstance(
        pure_dict['dict']['field_ref1'], config_dict.FieldReference
    )
    self.assertNotIsInstance(
        pure_dict['dict']['field_ref2'], config_dict.FieldReference
    )
    self.assertEqual(pure_dict['dict']['field_ref1'], field.get())
    self.assertEqual(pure_dict['dict']['field_ref2'], field.get())

    pure_dict_with_refs = cfg.to_dict(preserve_field_references=True)
    self.assertEqual(type(pure_dict_with_refs), dict)
    self.assertEqual(type(pure_dict_with_refs['dict']), dict)
    self.assertIsInstance(
        pure_dict_with_refs['dict']['field_ref1'], config_dict.FieldReference
    )
    self.assertIsInstance(
        pure_dict_with_refs['dict']['field_ref2'], config_dict.FieldReference
    )
    self.assertIs(
        pure_dict_with_refs['dict']['field_ref1'],
        pure_dict_with_refs['dict']['field_ref2'],
    )

    # Ensure FieldReferences in the dict are not the same as the FieldReferences
    # in the original ConfigDict.
    self.assertIsNot(
        pure_dict_with_refs['dict']['field_ref1'], cfg.dict['field_ref1']
    )

  def testToDictTypeUnsafe(self):
    """Tests interaction between ignore_type() and to_dict()."""
    cfg = config_dict.ConfigDict()
    cfg.string = config_dict.FieldReference(None, field_type=str)

    with cfg.ignore_type():
      cfg.string = 1
    self.assertEqual(1, cfg.to_dict(preserve_field_references=True)['string'])

  def testCopyAndResolveReferences(self):
    """Tests the .copy_and_resolve_references() method."""
    cfg = config_dict.ConfigDict()

    field = config_dict.FieldReference('a string')
    int_field = config_dict.FieldReference(5)
    cfg.dict = {
        'float': 2.3,
        'integer': 1,
        'field_ref1': field,
        'field_ref2': field,
        'field_ref_int1': int_field,
        'field_ref_int2': int_field + 5,
        'placeholder': config_dict.placeholder(str),
        'cfg': config_dict.ConfigDict({'integer': 1, 'int_field': int_field}),
    }

    cfg.ref = cfg.dict
    cfg.self_ref = cfg

    cfg_resolved = cfg.copy_and_resolve_references()
    for field, value in [
        ('float', 2.3),
        ('integer', 1),
        ('field_ref1', 'a string'),
        ('field_ref2', 'a string'),
        ('field_ref_int1', 5),
        ('field_ref_int2', 10),
        ('placeholder', None),
    ]:
      self.assertEqual(getattr(cfg_resolved.dict, field), value)
    for field, value in [('integer', 1), ('int_field', 5)]:
      self.assertEqual(getattr(cfg_resolved.dict.cfg, field), value)
    self.assertIs(cfg_resolved, cfg_resolved['self_ref'])
    self.assertIs(cfg_resolved['dict'], cfg_resolved['ref'])

  def testCopyAndResolveReferencesConfigTypes(self):
    """Tests that .copy_and_resolve_references() handles special types."""
    cfg_type_safe = config_dict.ConfigDict()
    int_field = config_dict.FieldReference(5)
    cfg_type_safe.field_ref1 = int_field
    cfg_type_safe.field_ref2 = int_field + 5

    cfg_type_safe.lock()
    cfg_type_safe_locked_resolved = cfg_type_safe.copy_and_resolve_references()
    self.assertTrue(cfg_type_safe_locked_resolved.is_locked)
    self.assertTrue(cfg_type_safe_locked_resolved.is_type_safe)

    cfg = config_dict.ConfigDict(type_safe=False)
    cfg.field_ref1 = int_field
    cfg.field_ref2 = int_field + 5

    cfg_resolved = cfg.copy_and_resolve_references()
    self.assertFalse(cfg_resolved.is_locked)
    self.assertFalse(cfg_resolved.is_type_safe)

    cfg.lock()
    cfg_locked_resolved = cfg.copy_and_resolve_references()
    self.assertTrue(cfg_locked_resolved.is_locked)
    self.assertFalse(cfg_locked_resolved.is_type_safe)

    for resolved in [
        cfg_type_safe_locked_resolved,
        cfg_resolved,
        cfg_locked_resolved,
    ]:
      self.assertEqual(resolved.field_ref1, 5)
      self.assertEqual(resolved.field_ref2, 10)

    frozen_cfg = config_dict.FrozenConfigDict(cfg_type_safe)
    frozen_cfg_resolved = frozen_cfg.copy_and_resolve_references()

    for resolved in [frozen_cfg, frozen_cfg_resolved]:
      self.assertEqual(resolved.field_ref1, 5)
      self.assertEqual(resolved.field_ref2, 10)
      self.assertIsInstance(resolved, config_dict.FrozenConfigDict)

  def testInitConfigDict(self):
    """Tests initializing a ConfigDict on a ConfigDict."""
    cfg = _get_test_config_dict()
    cfg_2 = config_dict.ConfigDict(cfg)
    self.assertIsNot(cfg_2, cfg)
    self.assertIs(cfg_2.float, cfg.float)
    self.assertIs(cfg_2.dict, cfg.dict)

    # Ensure ConfigDict fields are initialized as is
    dict_with_cfg_field = {'cfg': cfg}
    cfg_3 = config_dict.ConfigDict(dict_with_cfg_field)
    self.assertIs(cfg_3.cfg, cfg)

    # Now ensure it works with locking and type_safe
    cfg_4 = config_dict.ConfigDict(cfg, type_safe=False)
    cfg_4.lock()
    self.assertEqual(cfg_4, config_dict.ConfigDict(cfg_4))

  def testInitReferenceStructure(self):
    """Ensures initialization preserves reference structure."""
    x = [1, 2, 3]
    self_ref_dict = {
        'float': 2.34,
        'test_dict_1': _TEST_DICT,
        'test_dict_2': _TEST_DICT,
        'list': x,
    }
    self_ref_dict['self'] = self_ref_dict
    self_ref_dict['self_fr'] = config_dict.FieldReference(self_ref_dict)

    self_ref_cd = config_dict.ConfigDict(self_ref_dict)
    self.assertIs(self_ref_cd.test_dict_1, self_ref_cd.test_dict_2)
    self.assertIs(self_ref_cd, self_ref_cd.self)
    self.assertIs(self_ref_cd, self_ref_cd.self_fr)
    self.assertIs(self_ref_cd.list, x)
    self.assertEqual(self_ref_cd, self_ref_cd.self)

    self_ref_cd.self.int = 1
    self.assertEqual(self_ref_cd.int, 1)

    self_ref_cd_2 = config_dict.ConfigDict(self_ref_cd)
    self.assertIsNot(self_ref_cd_2, self_ref_cd)
    self.assertIs(self_ref_cd_2.self, self_ref_cd_2)
    self.assertIs(self_ref_cd_2.test_dict_1, self_ref_cd.test_dict_1)

  def testInitFieldReference(self):
    """Tests initialization with FieldReferences."""
    test_dict = dict(x=1, y=1)

    # Reference to a dict.
    reference = config_dict.FieldReference(test_dict)
    cfg = config_dict.ConfigDict()
    cfg.reference = reference
    self.assertIsInstance(cfg.reference, config_dict.ConfigDict)
    self.assertEqual(test_dict['x'], cfg.reference.x)
    self.assertEqual(test_dict['y'], cfg.reference.y)

    # Reference to a ConfigDict.
    test_configdict = config_dict.ConfigDict(test_dict)
    reference = config_dict.FieldReference(test_configdict)
    cfg = config_dict.ConfigDict()
    cfg.reference = reference

    test_configdict.x = 2
    self.assertEqual(test_configdict.x, cfg.reference.x)
    self.assertEqual(test_configdict.y, cfg.reference.y)

    # Reference to a reference.
    reference_int = config_dict.FieldReference(0)
    reference = config_dict.FieldReference(reference_int)
    cfg = config_dict.ConfigDict()
    cfg.reference = reference

    reference_int.set(1)
    self.assertEqual(reference_int.get(), cfg.reference)

  def testDeletingFields(self):
    """Tests whether it is possible to delete fields."""
    cfg = config_dict.ConfigDict()

    cfg.field1 = 123
    cfg.field2 = 123
    self.assertIn('field1', cfg)
    self.assertIn('field2', cfg)

    del cfg.field1

    self.assertNotIn('field1', cfg)
    self.assertIn('field2', cfg)

    del cfg.field2

    self.assertNotIn('field2', cfg)

    with self.assertRaises(AttributeError):
      del cfg.keys

    with self.assertRaises(KeyError):
      del cfg['keys']

  def testDeletingNestedFields(self):
    """Tests whether it is possible to delete nested fields."""

    cfg = config_dict.ConfigDict({
        'a': {
            'aa': [1, 2],
        },
        'b': {
            'ba': {
                'baa': 2,
                'bab': 3,
            },
            'bb': {1, 2, 3},
        },
    })

    self.assertIn('a', cfg)
    self.assertIn('aa', cfg.a)
    self.assertIn('baa', cfg.b.ba)

    del cfg['a.aa']
    self.assertIn('a', cfg)
    self.assertNotIn('aa', cfg.a)

    del cfg['a']
    self.assertNotIn('a', cfg)

    del cfg['b.ba.baa']
    self.assertIn('ba', cfg.b)
    self.assertIn('bab', cfg.b.ba)
    self.assertNotIn('baa', cfg.b.ba)

    del cfg['b.ba']
    self.assertNotIn('ba', cfg.b)
    self.assertIn('bb', cfg.b)

    with self.assertRaises(AttributeError):
      del cfg.keys

    with self.assertRaises(KeyError):
      del cfg['keys']

  def testSetAttr(self):
    """Tests whether it is possible to override an attribute."""
    cfg = config_dict.ConfigDict()
    with self.assertRaises(AttributeError):
      cfg.__setattr__('__class__', 'abc')

  def testPickling(self):
    """Tests whether ConfigDict can be pickled and unpickled."""
    cfg = _get_test_config_dict()
    cfg.lock()
    pickle_cfg = pickle.loads(pickle.dumps(cfg))
    self.assertTrue(pickle_cfg.is_locked)
    self.assertIsInstance(pickle_cfg, config_dict.ConfigDict)
    self.assertEqual(str(cfg), str(pickle_cfg))

  def testPicklingLockedWithCircularReference(self):
    cfg = config_dict.ConfigDict()
    cfg.a = 1.0
    cfg.b = config_dict.ConfigDict()
    cfg.b.c = config_dict.ConfigDict()
    cfg.b.d = cfg
    cfg.lock()

    pickle_cfg = pickle.loads(pickle.dumps(cfg))
    self.assertTrue(pickle_cfg.is_locked)
    self.assertTrue(pickle_cfg.b.is_locked)
    self.assertTrue(pickle_cfg.b.c.is_locked)
    self.assertTrue(pickle_cfg.b.d.is_locked)
    self.assertEqual(pickle_cfg.a, 1.0)
    self.assertIsInstance(pickle_cfg, config_dict.ConfigDict)
    self.assertIsInstance(pickle_cfg.b.d, config_dict.ConfigDict)

  def testPlaceholder(self):
    """Tests whether FieldReference works correctly as a placeholder."""
    cfg_element = config_dict.FieldReference(0)
    cfg = config_dict.ConfigDict(
        {'element': cfg_element, 'nested': {'element': cfg_element}}
    )

    # Type mismatch.
    with self.assertRaises(TypeError):
      cfg.element = 'string'

    cfg.element = 1
    self.assertEqual(cfg.element, cfg.nested.element)

  def testOptional(self):
    """Tests whether FieldReference works correctly as an optional field."""
    # Type mismatch at construction.
    with self.assertRaises(TypeError):
      config_dict.FieldReference(0, field_type=str)

    # None default and field_type.
    with self.assertRaises(ValueError):
      config_dict.FieldReference(None)

    cfg = config_dict.ConfigDict({
        'default': config_dict.FieldReference(0),
    })

    cfg.default = 1
    self.assertEqual(cfg.default, 1)

  def testOptionalNoDefault(self):
    """Tests optional field with no default value."""
    cfg = config_dict.ConfigDict({
        'nodefault': config_dict.FieldReference(None, field_type=str),
    })

    # Type mismatch with field with no default value.
    with self.assertRaises(TypeError):
      cfg.nodefault = 1

    cfg.nodefault = 'string'
    self.assertEqual(cfg.nodefault, 'string')

  def testGetType(self):
    """Tests whether types are correct for FieldReference fields."""
    cfg = config_dict.ConfigDict()
    cfg.integer = 123
    cfg.ref = config_dict.FieldReference(123)
    cfg.ref_nodefault = config_dict.FieldReference(None, field_type=int)

    self.assertEqual(cfg.get_type('integer'), int)
    self.assertEqual(cfg.get_type('ref'), int)
    self.assertEqual(cfg.get_type('ref_nodefault'), int)

    # Check errors in case of misspelled key.
    with self.assertRaisesRegex(AttributeError, 'Did you.*ref_nodefault.*'):
      cfg.get_type('ref_nodefualt')

    with self.assertRaisesRegex(AttributeError, 'Did you.*integer.*'):
      cfg.get_type('integre')

  def testFieldRefConvertDictFalse(self):
    config = config_dict.ConfigDict(
        initial_dictionary={'a': {}},
        convert_dict=False,
    )
    config.b = config.get_ref('a')
    self.assertIsInstance(config.b, dict)


class ConfigDictUpdateTest(absltest.TestCase):

  def testUpdateSimple(self):
    """Tests updating from one ConfigDict to another."""
    first = config_dict.ConfigDict()
    first.x = 5
    first.y = 'truman'
    first.q = 2.0
    second = config_dict.ConfigDict()
    second.x = 9
    second.y = 'wilson'
    second.z = 'washington'
    first.update(second)

    self.assertEqual(first.x, 9)
    self.assertEqual(first.y, 'wilson')
    self.assertEqual(first.z, 'washington')
    self.assertEqual(first.q, 2.0)

  def testUpdateNothing(self):
    """Tests updating a ConfigDict with no arguments."""
    cfg = config_dict.ConfigDict()
    cfg.x = 5
    cfg.y = 9
    cfg.update()
    self.assertLen(cfg, 2)
    self.assertEqual(cfg.x, 5)
    self.assertEqual(cfg.y, 9)

  def testUpdateFromDict(self):
    """Tests updating a ConfigDict from a dict."""
    cfg = config_dict.ConfigDict()
    cfg.x = 5
    cfg.y = 9
    cfg.update({'x': 6, 'z': 2})
    self.assertEqual(cfg.x, 6)
    self.assertEqual(cfg.y, 9)
    self.assertEqual(cfg.z, 2)

  def testUpdateFromKwargs(self):
    """Tests updating a ConfigDict from kwargs."""
    cfg = config_dict.ConfigDict()
    cfg.x = 5
    cfg.y = 9
    cfg.update(x=6, z=2)
    self.assertEqual(cfg.x, 6)
    self.assertEqual(cfg.y, 9)
    self.assertEqual(cfg.z, 2)

  def testUpdateFromDictAndKwargs(self):
    """Tests updating a ConfigDict from a dict and kwargs."""
    cfg = config_dict.ConfigDict()
    cfg.x = 5
    cfg.y = 9
    cfg.update({'x': 4, 'z': 2}, x=6)
    self.assertEqual(cfg.x, 6)  # kwarg overrides value from dict
    self.assertEqual(cfg.y, 9)
    self.assertEqual(cfg.z, 2)

  def testUpdateFromMultipleDictTypeError(self):
    """Tests that updating a ConfigDict from two dicts raises a TypeError."""
    cfg = config_dict.ConfigDict()
    cfg.x = 5
    cfg.y = 9
    with self.assertRaisesRegex(
        TypeError, 'update expected at most 1 arguments, got 2'
    ):
      cfg.update({'x': 4}, {'z': 2})

  def testUpdateNested(self):
    """Tests updating a ConfigDict from a nested dict."""
    cfg = config_dict.ConfigDict()
    cfg.subcfg = config_dict.ConfigDict()
    cfg.p = 5
    cfg.q = 6
    cfg.subcfg.y = 9
    cfg.update({'p': 4, 'subcfg': {'y': 10, 'z': 5}})
    self.assertEqual(cfg.p, 4)
    self.assertEqual(cfg.q, 6)
    self.assertEqual(cfg.subcfg.y, 10)
    self.assertEqual(cfg.subcfg.z, 5)

  def _assert_associated(self, cfg1, cfg2, key):
    self.assertEqual(cfg1[key], cfg2[key])
    cfg1[key] = 1
    cfg2[key] = 2
    self.assertEqual(cfg1[key], 2)
    cfg1[key] = 3
    self.assertEqual(cfg2[key], 3)

  def testUpdateFieldReference(self):
    """Tests updating to/from FieldReference fields."""
    # Updating FieldReference...
    ref = config_dict.FieldReference(1)
    cfg = config_dict.ConfigDict(dict(a=ref, b=ref))
    # from value.
    cfg.update(config_dict.ConfigDict(dict(a=2)))
    self.assertEqual(cfg.a, 2)
    self.assertEqual(cfg.b, 2)
    # from FieldReference.
    error_message = 'Cannot update a FieldReference from another FieldReference'
    with self.assertRaisesRegex(TypeError, error_message):
      cfg.update(config_dict.ConfigDict(dict(a=config_dict.FieldReference(2))))
    with self.assertRaisesRegex(TypeError, error_message):
      cfg.update(config_dict.ConfigDict(dict(b=config_dict.FieldReference(2))))

    # Updating empty ConfigDict with FieldReferences.
    ref = config_dict.FieldReference(1)
    cfg_from = config_dict.ConfigDict(dict(a=ref, b=ref))
    cfg = config_dict.ConfigDict()
    cfg.update(cfg_from)
    self._assert_associated(cfg, cfg_from, 'a')
    self._assert_associated(cfg, cfg_from, 'b')

    # Updating values with FieldReferences.
    ref = config_dict.FieldReference(1)
    cfg_from = config_dict.ConfigDict(dict(a=ref, b=ref))
    cfg = config_dict.ConfigDict(dict(a=2, b=3))
    cfg.update(cfg_from)
    self._assert_associated(cfg, cfg_from, 'a')
    self._assert_associated(cfg, cfg_from, 'b')

  def testUpdateFromFlattened(self):
    cfg = config_dict.ConfigDict({'a': 1, 'b': {'c': {'d': 2}}})
    updates = {'a': 2, 'b.c.d': 3, 'b.c.e': 4}
    cfg.update_from_flattened_dict(updates)
    self.assertEqual(cfg.a, 2)
    self.assertEqual(cfg.b.c.d, 3)
    self.assertEqual(cfg.b.c.e, 4)

  def testUpdateFromFlattenedWithPrefix(self):
    cfg = config_dict.ConfigDict({'a': 1, 'b': {'c': {'d': 2}}})
    updates = {'a': 2, 'b.c.d': 3}
    cfg.b.update_from_flattened_dict(updates, 'b.')
    self.assertEqual(cfg.a, 1)
    self.assertEqual(cfg.b.c.d, 3)

  def testUpdateFromFlattenedNotFound(self):
    cfg = config_dict.ConfigDict({'a': 1, 'b': {'c': {'d': 2}}})
    updates = {'a': 2, 'b.d.e': 3}
    with self.assertRaisesRegex(
        KeyError, 'Key "b.d.e" cannot be set as "b.d" was not found.'
    ):
      cfg.update_from_flattened_dict(updates)

  def testUpdateFromFlattenedWrongType(self):
    cfg = config_dict.ConfigDict({'a': 1, 'b': {'c': {'d': 2}}})
    updates = {'a.b.c': 2}
    with self.assertRaisesRegex(
        KeyError, 'Key "a.b.c" cannot be updated as "a" is not a ConfigDict.'
    ):
      cfg.update_from_flattened_dict(updates)

  def testUpdateFromFlattenedTupleListConversion(self):
    cfg = config_dict.ConfigDict({
        'a': 1,
        'b': {
            'c': {
                'd': (1, 2, 3, 4, 5),
            }
        },
    })
    updates = {
        'b.c.d': [2, 4, 6, 8],
    }
    cfg.update_from_flattened_dict(updates)
    self.assertIsInstance(cfg.b.c.d, tuple)
    self.assertEqual(cfg.b.c.d, (2, 4, 6, 8))

  def testUpdateFromFlattenedTupleListIndexConversion(self):
    cfg = config_dict.ConfigDict({
        'a': 1,
        'b': {
            'c': {
                'd': (1, 2, 3, 4, 5),
            },
            'e': [
                config_dict.ConfigDict({
                    'f': 4,
                    'g': 5,
                }),
                config_dict.ConfigDict({
                    'f': 6,
                    'g': 7,
                }),
            ],
        },
    })
    updates = {
        'b.c.d[2]': 9,
        'b.e[1].f': 12,
    }
    cfg.update_from_flattened_dict(updates)
    self.assertIsInstance(cfg.b.c.d, tuple)
    self.assertEqual(cfg.b.c.d, (1, 2, 9, 4, 5))
    self.assertCountEqual(cfg.b.e[1], {'f': 12, 'g': 7})

  def testDecodeError(self):
    # ConfigDict containing two strings with incompatible encodings.
    cfg = config_dict.ConfigDict({
        'dill': pickle.dumps(_test_function, protocol=pickle.HIGHEST_PROTOCOL),
        'unicode': 'unicode string',
    })

    with self.assertRaises(TypeError):
      cfg.to_json()

  def testConvertDict(self):
    """Test automatic conversion, or not, of dict to ConfigDict."""
    cfg = config_dict.ConfigDict()
    cfg.a = dict(b=dict(c=0))
    self.assertIsInstance(cfg.a, config_dict.ConfigDict)
    self.assertIsInstance(cfg.a.b, config_dict.ConfigDict)

    cfg = config_dict.ConfigDict(convert_dict=False)
    cfg.a = dict(b=dict(c=0))
    self.assertNotIsInstance(cfg.a, config_dict.ConfigDict)
    self.assertIsInstance(cfg.a, dict)
    self.assertIsInstance(cfg.a['b'], dict)

  def testConvertDictInInitialValue(self):
    """Test automatic conversion, or not, of dict to ConfigDict."""
    initial_dict = dict(a=dict(b=dict(c=0)))
    cfg = config_dict.ConfigDict(initial_dict)
    self.assertIsInstance(cfg.a, config_dict.ConfigDict)
    self.assertIsInstance(cfg.a.b, config_dict.ConfigDict)

    cfg = config_dict.ConfigDict(initial_dict, convert_dict=False)
    self.assertNotIsInstance(cfg.a, config_dict.ConfigDict)
    self.assertIsInstance(cfg.a, dict)
    self.assertIsInstance(cfg.a['b'], dict)

  def testConvertDictInCopyAndResolveReferences(self):
    """Test conversion, or not, of dict in copy and resolve references."""
    cfg = config_dict.ConfigDict()
    cfg.a = dict(b=dict(c=0))
    copied_cfg = cfg.copy_and_resolve_references()
    self.assertIsInstance(copied_cfg.a, config_dict.ConfigDict)
    self.assertIsInstance(copied_cfg.a.b, config_dict.ConfigDict)

    cfg = config_dict.ConfigDict(convert_dict=False)
    cfg.a = dict(b=dict(c=0))
    copied_cfg = cfg.copy_and_resolve_references()
    self.assertNotIsInstance(copied_cfg.a, config_dict.ConfigDict)
    self.assertIsInstance(copied_cfg.a, dict)
    self.assertIsInstance(copied_cfg.a['b'], dict)

  def testConvertDictTypeCompat(self):
    """Test that automatic conversion to ConfigDict doesn't trigger type errors."""
    cfg = config_dict.ConfigDict()
    cfg.a = {}
    self.assertIsInstance(cfg.a, config_dict.ConfigDict)
    # This checks that dict to configdict casting doesn't produce type mismatch.
    cfg.a = {}

  def testYamlNoConvert(self):
    """Test deserialisation from YAML without convert dict.

    This checks backward compatibility of deserialisation.
    """
    cfg = config_dict.ConfigDict(dict(a=1))
    self.assertTrue(yaml.load(cfg.to_yaml(), yaml.UnsafeLoader)._convert_dict)

  def testRecursiveRename(self):
    """Test recursive_rename.

    The dictionary should be the same but with the specified name changed.
    """
    cfg = config_dict.ConfigDict(_TEST_DICT)
    new_cfg = config_dict.recursive_rename(cfg, 'float', 'double')
    # Check that the new config has float changed to double as we expect
    self.assertEqual(new_cfg.to_dict(), _TEST_DICT_CHANGE_FLOAT_NAME)
    # Check that the original config is unchanged
    self.assertEqual(cfg.to_dict(), _TEST_DICT)

  def testGetOnewayRef(self):
    cfg = config_dict.create(a=1)
    cfg.b = cfg.get_oneway_ref('a')

    cfg.a = 2
    self.assertEqual(2, cfg.b)

    cfg.b = 3
    self.assertEqual(2, cfg.a)
    self.assertEqual(3, cfg.b)

  def testNoSortKey(self):
    cfg = config_dict.ConfigDict({'z': 1, 'a': 2}, sort_keys=False)
    self.assertEqual(list(cfg.keys()), ['z', 'a'])

    cfg = config_dict.ConfigDict({'z': 1, 'a': 2})
    self.assertEqual(list(cfg.keys()), ['a', 'z'])

  def testDotInKey(self):
    cfg = config_dict.ConfigDict(allow_dotted_keys=True)
    cfg['a.b.c'] = 123
    self.assertEqual(cfg['a.b.c'], 123)

    with self.assertRaises(AttributeError):
      _ = cfg.a

    self.assertEqual(cfg.to_dict(), {'a.b.c': 123})

    del cfg['a.b.c']
    self.assertEqual(cfg.to_dict(), {})

  def testIntorFloatKey(self):
    cfg = config_dict.ConfigDict({1: 1})
    cfg[1] = 2
    self.assertEqual(cfg[1], 2)
    self.assertEqual(cfg.to_dict(), {1: 2})

    del cfg[1]
    self.assertEqual(cfg.to_dict(), {})

    cfg = config_dict.ConfigDict({1.1: 1})
    cfg[1.1] = 2
    self.assertEqual(cfg[1.1], 2)

    with self.assertRaises(KeyError):
      _ = cfg[1]
    self.assertEqual(cfg.to_dict(), {1.1: 2})

    del cfg[1.1]
    self.assertEqual(cfg.to_dict(), {})


class CreateTest(absltest.TestCase):

  def testBasic(self):
    config = config_dict.create(a=1, b='b')
    dct = {'a': 1, 'b': 'b'}
    self.assertEqual(config.to_dict(), dct)

  def testNested(self):
    config = config_dict.create(
        data=config_dict.create(game='freeway'),
        model=config_dict.create(num_hidden=1000),
    )

    dct = {'data': {'game': 'freeway'}, 'model': {'num_hidden': 1000}}
    self.assertEqual(config.to_dict(), dct)


class PlaceholderTest(absltest.TestCase):

  def testBasic(self):
    config = config_dict.create(a=1, b=config_dict.placeholder(int))
    self.assertEqual(config.to_dict(), {'a': 1, 'b': None})
    config.b = 5
    self.assertEqual(config.to_dict(), {'a': 1, 'b': 5})

  def testTypeChecking(self):
    config = config_dict.create(a=1, b=config_dict.placeholder(int))
    with self.assertRaises(TypeError):
      config.b = 'chutney'

  def testRequired(self):
    config = config_dict.create(a=config_dict.required_placeholder(int))
    ref = config.get_ref('a')
    with self.assertRaises(config_dict.RequiredValueError):
      config.a  # pylint: disable=pointless-statement
    with self.assertRaises(config_dict.RequiredValueError):
      config.to_dict()
    with self.assertRaises(config_dict.RequiredValueError):
      ref.get()

    config.a = 10
    self.assertEqual(config.to_dict(), {'a': 10})
    self.assertEqual(str(config), yaml.dump({'a': 10}))

    # Reset to None and check we still get an error.
    config.a = None
    with self.assertRaises(config_dict.RequiredValueError):
      config.a  # pylint: disable=pointless-statement

    # Set to a different value using the reference obtained calling get_ref().
    ref.set(5)
    self.assertEqual(config.to_dict(), {'a': 5})
    self.assertEqual(str(config), yaml.dump({'a': 5}))

    # dict placeholder.
    test_dict = {'field': 10}
    config = config_dict.create(
        a=config_dict.required_placeholder(dict),
        b=config_dict.FieldReference(test_dict.copy()),
    )
    # ConfigDict initialization converts dict to ConfigDict.
    self.assertEqual(test_dict, config.b.to_dict())
    config.a = test_dict
    self.assertEqual(test_dict, config.a)


class CycleTest(absltest.TestCase):

  def testCycle(self):
    config = config_dict.create(a=1)
    config.b = config.get_ref('a') + config.get_ref('a')
    self.assertFalse(config.get_ref('b').has_cycle())
    with self.assertRaises(config_dict.MutabilityError):
      config.a = config.get_ref('a')
    with self.assertRaises(config_dict.MutabilityError):
      config.a = config.get_ref('b')


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for ConfigDict examples.

Ensures that config_dict_basic, config_dict_initialization, config_dict_lock,
config_dict_placeholder, field_reference, frozen_config_dict run successfully.
"""

from absl.testing import absltest
from absl.testing import parameterized
from ml_collections.config_dict.examples import config_dict_advanced
from ml_collections.config_dict.examples import config_dict_basic
from ml_collections.config_dict.examples import config_dict_initialization
from ml_collections.config_dict.examples import config_dict_lock
from ml_collections.config_dict.examples import config_dict_placeholder
from ml_collections.config_dict.examples import field_reference
from ml_collections.config_dict.examples import frozen_config_dict


class ConfigDictExamplesTest(parameterized.TestCase):

  @parameterized.parameters(config_dict_advanced, config_dict_basic,
                            config_dict_initialization, config_dict_lock,
                            config_dict_placeholder, field_reference,
                            frozen_config_dict)
  def testScriptRuns(self, example_name):
    example_name.main(None)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example of basic FrozenConfigDict usage.

This example shows the most basic usage of FrozenConfigDict, highlighting
the differences between FrozenConfigDict and ConfigDict and including
converting between the two.
"""

from absl import app
from ml_collections import config_dict


def print_section(name):
  print()
  print()
  print('-' * len(name))
  print(name.upper())
  print('-' * len(name))
  print()


def main(_):
  print_section('Attribute Types.')
  cfg = config_dict.ConfigDict()
  cfg.int = 1
  cfg.list = [1, 2, 3]
  cfg.tuple = (1, 2, 3)
  cfg.set = {1, 2, 3}
  cfg.frozenset = frozenset({1, 2, 3})
  cfg.dict = {
      'nested_int': 4,
      'nested_list': [4, 5, 6],
      'nested_tuple': ([4], 5, 6),
  }

  print('Types of cfg fields:')
  print('list: ', type(cfg.list))  # List
  print('set: ', type(cfg.set))  # Set
  print('nested_list: ', type(cfg.dict.nested_list))  # List
  print('nested_tuple[0]: ', type(cfg.dict.nested_tuple[0]))  # List

  frozen_cfg = config_dict.FrozenConfigDict(cfg)
  print('\nTypes of FrozenConfigDict(cfg) fields:')
  print('list: ', type(frozen_cfg.list))  # Tuple
  print('set: ', type(frozen_cfg.set))  # Frozenset
  print('nested_list: ', type(frozen_cfg.dict.nested_list))  # Tuple
  print('nested_tuple[0]: ', type(frozen_cfg.dict.nested_tuple[0]))  # Tuple

  cfg_from_frozen = config_dict.ConfigDict(frozen_cfg)
  print('\nTypes of ConfigDict(FrozenConfigDict(cfg)) fields:')
  print('list: ', type(cfg_from_frozen.list))  # List
  print('set: ', type(cfg_from_frozen.set))  # Set
  print('nested_list: ', type(cfg_from_frozen.dict.nested_list))  # List
  print('nested_tuple[0]: ', type(cfg_from_frozen.dict.nested_tuple[0]))  # List

  print('\nCan use FrozenConfigDict.as_configdict() to convert to ConfigDict:')
  print(cfg_from_frozen == frozen_cfg.as_configdict())  # True

  print_section('Immutability.')
  try:
    frozen_cfg.new_field = 1  # Raises AttributeError because of immutability.
  except AttributeError as e:
    print(e)

  print_section('"==" and eq_as_configdict().')
  # FrozenConfigDict.__eq__() is not type-invariant with respect to ConfigDict
  print(frozen_cfg == cfg)  # False
  # FrozenConfigDict.eq_as_configdict() is type-invariant with respect to
  # ConfigDict
  print(frozen_cfg.eq_as_configdict(cfg))  # True
  # .eq_as_congfigdict() is also a method of ConfigDict
  print(cfg.eq_as_configdict(frozen_cfg))  # True


if __name__ == '__main__':
  app.run(main)



# Copyright 2024 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example of FieldReference usage.

This shows how to use FieldReferences for lazy computation.
"""

from absl import app
from ml_collections import config_dict


def lazy_computation():
  """Simple example of lazy computation with `configdict.FieldReference`."""
  ref = config_dict.FieldReference(1)
  print(ref.get())  # Prints 1

  add_ten = ref.get() + 10  # ref.get() is an integer and so is add_ten
  add_ten_lazy = ref + 10  # add_ten_lazy is a FieldReference - NOT an integer

  print(add_ten)  # Prints 11
  print(add_ten_lazy.get())  # Prints 11 because ref's value is 1

  # Addition is lazily computed for FieldReferences so changing ref will change
  # the value that is used to compute add_ten.
  ref.set(5)
  print(add_ten)  # Prints 11
  print(add_ten_lazy.get())  # Prints 15 because ref's value is 5


def change_lazy_computation():
  """Overriding lazily computed values."""
  config = config_dict.ConfigDict()
  config.reference = 1
  config.reference_0 = config.get_ref('reference') + 10
  config.reference_1 = config.get_ref('reference') + 20
  config.reference_1_0 = config.get_ref('reference_1') + 100

  print(config.reference)  # Prints 1.
  print(config.reference_0)  # Prints 11.
  print(config.reference_1)  # Prints 21.
  print(config.reference_1_0)  # Prints 121.

  config.reference_1 = 30

  print(config.reference)  # Prints 1 (unchanged).
  print(config.reference_0)  # Prints 11 (unchanged).
  print(config.reference_1)  # Prints 30.
  print(config.reference_1_0)  # Prints 130.


def create_cycle():
  """Creates a cycle within a ConfigDict."""
  config = config_dict.ConfigDict()
  config.integer_field = 1
  config.bigger_integer_field = config.get_ref('integer_field') + 10

  try:
    # Raises a MutabilityError because setting config.integer_field would
    # cause a cycle.
    config.integer_field = config.get_ref('bigger_integer_field') + 2
  except config_dict.MutabilityError as e:
    print(e)


def lazy_configdict():
  """Example usage of lazy computation with ConfigDict."""
  config = config_dict.ConfigDict()
  config.reference_field = config_dict.FieldReference(1)
  config.integer_field = 2
  config.float_field = 2.5

  # No lazy evaluatuations because we didn't use get_ref()
  config.no_lazy = config.integer_field * config.float_field

  # This will lazily evaluate ONLY config.integer_field
  config.lazy_integer = config.get_ref('integer_field') * config.float_field

  # This will lazily evaluate ONLY config.float_field
  config.lazy_float = config.integer_field * config.get_ref('float_field')

  # This will lazily evaluate BOTH config.integer_field and config.float_Field
  config.lazy_both = (config.get_ref('integer_field') *
                      config.get_ref('float_field'))

  config.integer_field = 3
  print(config.no_lazy)  # Prints 5.0 - It uses integer_field's original value

  print(config.lazy_integer)  # Prints 7.5

  config.float_field = 3.5
  print(config.lazy_float)  # Prints 7.0
  print(config.lazy_both)  # Prints 10.5


def lazy_configdict_advanced():
  """Advanced lazy computation with ConfigDict."""
  # FieldReferences can be used with ConfigDict as well
  config = config_dict.ConfigDict()
  config.float_field = 12.6
  config.integer_field = 123
  config.list_field = [0, 1, 2]

  config.float_multiply_field = config.get_ref('float_field') * 3
  print(config.float_multiply_field)  # Prints 37.8

  config.float_field = 10.0
  print(config.float_multiply_field)  # Prints 30.0

  config.longer_list_field = config.get_ref('list_field') + [3, 4, 5]
  print(config.longer_list_field)  # Prints [0, 1, 2, 3, 4, 5]

  config.list_field = [-1]
  print(config.longer_list_field)  # Prints [-1, 3, 4, 5]

  # Both operands can be references
  config.ref_subtraction = (
      config.get_ref('float_field') - config.get_ref('integer_field'))
  print(config.ref_subtraction)  # Prints -113.0

  config.integer_field = 10
  print(config.ref_subtraction)  # Prints 0.0


def main(argv=()):
  del argv  # Unused.
  lazy_computation()
  lazy_configdict()
  change_lazy_computation()
  create_cycle()
  lazy_configdict_advanced()


if __name__ == '__main__':
  app.run()



# Copyright 2024 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for spliting flag prefixes."""

import ast
import dataclasses as dc
import functools
import types
import typing
from typing import Any, MutableSequence, Optional, Sequence, Tuple, Union, Type

from ml_collections import config_dict


NoneType = type(None)


_AST_SPLIT_CONFIG_PATH = {
    ast.Attribute: lambda n: (*_split_node(n.value), n.attr),
    ast.Index: lambda i: _split_node(i.value),
    ast.Name: lambda n: (n.id,),
    ast.Slice: lambda i: slice(*map(_split_node, (i.lower, i.upper, i.step))),
    ast.Subscript: lambda n: (*_split_node(n.value), _split_node(n.slice)),
    type(None): lambda n: None
}


def _split_node(node):
  return _AST_SPLIT_CONFIG_PATH.get(type(node), ast.literal_eval)(node)


def split(config_path: str) -> Tuple[Any]:
  """Returns config_path split into a tuple of parts.

  Example usage:
    >>> assert config_path.split('a.b.cc') == ('a', 'b', 'cc')
    >>> assert config_path.split('a.b["cc.d"]') == ('a', 'b', 'cc.d')
    >>> assert config_path.split('a.b[10]') == ('a', 'b', 10)
    >>> assert config_path.split('a[(1, 2)]') == ('a', (1, 2))
    >>> assert config_path.split('a[:]') == ('a', slice(None))

  Args:
    config_path: Input path to be split - see example usage.

  Returns:
    Tuple of config_path split into parts. Parts are attributes or subscripts.
    Attrributes are treated as strings and subscripts are parsed using
    ast.literal_eval. It is up to the caller to ensure all returned types are
    valid.

  Raises:
    ValueError: Failed to parse config_path.
  """
  try:
    node = ast.parse(config_path, mode='eval')
  except SyntaxError as e:
    raise ValueError(f'Could not parse {config_path!r}: {e!r}') from None
  if isinstance(node, ast.Expression):
    result = _split_node(node.body)
    if isinstance(result, tuple):
      return result
  raise ValueError(config_path)


def _get_item_or_attribute(config, field,
                           field_path: Optional[str] = None):
  """Returns attribute of member failing that the item."""
  if isinstance(field, str) and hasattr(config, field):
    return getattr(config, field)
  if hasattr(config, '__getitem__'):
    return config[field]
  if isinstance(field, int):
    raise IndexError(
        f'{type(config)} does not support integer indexing [{field}]]. '
        f'Attempting to lookup: {field_path}')
  raise KeyError(
      f'Attribute {type(config)}.{field} does not exist '
      'and the type does not support indexing. '
      f'Attempting to lookup: {field_path}')


def _get_holder_field(config_path: str, config: Any) -> Tuple[Any, str]:
  """Returns the last part config_path and config to allow assignment.

  Example usage:
    >>> config = {'a': {'b', {'c', 10}}}
    >>> holder, lastfield = _get_holder_field('a.b.c', config)
    >>> assert lastfield == 'c'
    >>> assert holder is config['a']['b']
    >>> assert holder[lastfield] == 10

  Args:
    config_path: Any string that `split` can process.
    config: A nested datastructure that can be accessed via
      _get_item_or_attribute

  Returns:
    The penultimate object when walking config with config_path. And the final
    part of the config path.

  Raises:
    IndexError: Integer field not found in nested structure.
    KeyError: Non-integer field not found in nested structure.
    ValueError: Empty/invalid config_path after parsing.
  """
  fields = split(config_path)
  if not fields:
    raise ValueError('Path cannot be empty')
  get_item = functools.partial(_get_item_or_attribute, field_path=config_path)
  holder = functools.reduce(get_item, fields[:-1], config)
  return holder, fields[-1]


def get_value(config_path: str, config: Any):
  """Gets value of a single field.

  Example usage:
    >>> config = {'a': {'b', {'c', 10}}}
    >>> assert config_path.get_value('a.b.c', config) == 10

  Args:
    config_path: Any string that `split` can process.
    config: A nested datastructure

  Returns:
    The last object when walking config with config_path.

  Raises:
    IndexError: Integer field not found in nested structure.
    KeyError: Non-integer field not found in nested structure.
    ValueError: Empty/invalid config_path after parsing.
  """
  get_item = functools.partial(_get_item_or_attribute, field_path=config_path)
  return functools.reduce(get_item, split(config_path), config)


def initialize_missing_parent_fields(
    config: Any, override: str,
    allowed_missing: Sequence[str]):
  """Adds some missing nested holder fields for a particular override.

  For example if override is 'config.a.b.c' and config.a is None, it
  will default initialize config.a, and if config.a.b is None will default
  initialize it as well. Only overrides present in allowed_missing will
  be initialized.

  Args:
    config: config object (typically dataclass)
    override: dot joined override name.
    allowed_missing: list of overrides that are allowed
    to be set. For example, if override is 'a.b.c.d',
    allowed_missing could be ['a.b.c', 'a', 'foo.bar'].

  Raises:
    ValueError: if parent field is not of dataclass type.
  """
  fields = split(override)
  # Collect the tree levels at which we are alloed to create override
  allowed_levels = {len(split(x)) for x in allowed_missing if
                    override.startswith(x + '.')}
  child = config
  for level, f in enumerate(fields[:-1], 1):
    parent = child
    child = _get_item_or_attribute(parent, f, override)
    if child is not None:
      continue
    # Field is not yet present, see if we should create it instead.
    field_type = get_type(f, parent)
    # Note: these two assertions below are mostly guard
    # rails to prevent behaviors that might be confusing/accidental.
    # Specifically we disallow implicit creation of parent fields,
    # creating non dataclass objects. They can be revisited
    # in the future.
    if not dc.is_dataclass(field_type):
      raise ValueError(
          f'Override {override} can not be applied because '
          f'field "{f}" is None, and its type "{field_type}" is not a '
          f'dataclass in the parent of type "{type(parent)}".')

    if level not in allowed_levels:
      raise ValueError(
          f'Flag {override} can not be applied because '
          f'field "{f}" is None by default and it is not explicitly '
          'provided in flags (it can be default intialized by '
          f'providing --<path-to-{f}>.{f}=build flag')
    try:
      child = field_type()
    except Exception as e:
      raise ValueError(
          f'Override {override} can not be applied because '
          f'field "{f}" of type {field_type} can not be default instantiated:'
          f'{e}') from e
    set_value(f, parent, child)


def get_origin(type_spec: type) -> Optional[type]:   # pylint: disable=g-bare-generic drop when 3.7 support is not needed
  """Call typing.get_origin, with a fallback for Python 3.7 and below."""
  if hasattr(typing, 'get_origin'):
    return typing.get_origin(type_spec)
  return getattr(type_spec, '__origin__', None)


def get_args(type_spec: type) -> Union[NoneType, Tuple[type, ...]]:  # pylint: disable=g-bare-generic drop when 3.7 support is not needed
  """Call typing.get_args, with fallback for Python 3.7 and below."""
  if hasattr(typing, 'get_args'):
    return typing.get_args(type_spec)
  return getattr(type_spec, '__args__', NoneType)


def _is_union_type(type_spec: type) -> bool:  # pylint: disable=g-bare-generic drop when 3.7 support is not needed
  """Cheeck if a type_spec is a Union type or not."""
  # UnionType was only introduced in python 3.10. We need getattr for
  # backward compatibility.
  return get_origin(type_spec) in [Union, getattr(types, 'UnionType', Union)]


def extract_type_from_optional(type_spec: type) -> Optional[type]:  # pylint: disable=g-bare-generic drop when 3.7 support is not needed
  """If type_spec is of type Optional[T], returns T object, otherwise None"""
  if not _is_union_type(type_spec):
    return None
  non_none = [t for t in get_args(type_spec) if t is not NoneType]
  if len(non_none) != 1:
    return None
  return non_none[0]


def normalize_type(type_spec: type) -> type:  # pylint: disable=g-bare-generic drop when 3.7 support is not needed
  """Normalizes a type object.

  Strips all None types from the type specification and returns the remaining
  single type. This is primarily useful for Optional type annotations in which
  case it will strip out the NoneType and return the inner type.

  Args:
    type_spec: The type to normalize.

  Raises:
    TypeError: If there is not exactly 1 non-None type in the union.
  Returns:
    The normalized type.
  """
  if _is_union_type(type_spec):
    subtype = extract_type_from_optional(type_spec)
    if subtype is None:
      raise TypeError(f'Unable to normalize ambiguous type: {type_spec}')
    return subtype

  return type_spec


def get_type(
    config_path: str,
    config: Any,
    normalize=True,
    default_type: Optional[Type[Any]] = None,
):
  """Gets type of field in config described by a config_path.

  Example usage:
    >>> config = {'a': {'b', {'c', 10}}}
    >>> assert config_path.get_type('a.b.c', config) is int

  Args:
    config_path: Any string that `split` can process.
    config: A nested datastructure
    normalize: whether to normalize the type (in particular strip Optional
      annotations on dataclass fields)
    default_type: If the `config_path` is not found and `default_type` is set,
      the `default_type` is returned.

  Returns:
    The type of last object when walking config with config_path.

  Raises:
    IndexError: Integer field not found in nested structure.
    KeyError: Non-integer field not found in nested structure.
    ValueError: Empty/invalid config_path after parsing.
    TypeError: Ambiguous type annotation on dataclass field.
  """
  holder, field = _get_holder_field(config_path, config)
  # Check if config is a DM collection and hence has attribute get_type()
  if isinstance(holder,
                (config_dict.ConfigDict, config_dict.FieldReference)):
    if default_type is not None and field not in holder:
      return default_type
    return holder.get_type(field)
  # For dataclasses we can just use the type annotation.
  elif dc.is_dataclass(holder):
    matches = [f.type for f in dc.fields(holder) if f.name == field]
    if not matches:
      raise KeyError(f'Field {field} not found on dataclass {type(holder)}')
    return normalize_type(matches[0]) if normalize else matches[0]
  else:
    return type(_get_item_or_attribute(holder, field, config_path))


def is_optional(config_path: str, config: Any) -> bool:
  raw_type = get_type(config_path, config, normalize=False)
  return extract_type_from_optional(raw_type) is not None


def set_value(
    config_path: str,
    config: Any,
    value: Any,
    *,
    accept_new_attributes: bool = False,
):
  """Sets value of field described by config_path.

  Example usage:
    >>> config = {'a': {'b', {'c', 10}}}
    >>> config_path.set_value('a.b.c', config, 20)
    >>> assert config['a']['b']['c'] == 20

  Args:
    config_path: Any string that `split` can process.
    config: A nested datastructure
    value: A value to assign to final field.
    accept_new_attributes: If `True`, the new config attributes can be added

  Raises:
    IndexError: Integer field not found in nested structure.
    KeyError: Non-integer field not found in nested structure.
    ValueError: Empty/invalid config_path after parsing.
  """
  holder, field = _get_holder_field(config_path, config)

  if isinstance(field, int) and isinstance(holder, MutableSequence):
    holder[field] = value
  elif hasattr(holder, '__setitem__') and (
      field in holder or accept_new_attributes
  ):
    holder[field] = value
  elif hasattr(holder, str(field)):
    setattr(holder, str(field), value)
  else:
    if isinstance(field, int):
      raise IndexError(
          f'{field} is not a valid index for {type(holder)} '
          f'(in: {config_path})')
    raise KeyError(f'{field} is not a valid key or attribute of {type(holder)} '
                   f'(in: {config_path})')



# Copyright 2024 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example of initialization features and gotchas in a ConfigDict.
"""

import copy

from absl import app
from ml_collections import config_dict


def print_section(name):
  print()
  print()
  print('-' * len(name))
  print(name.upper())
  print('-' * len(name))
  print()


def main(_):

  inner_dict = {'list': [1, 2], 'tuple': (1, 2, [3, 4], (5, 6))}
  example_dict = {
      'string': 'tom',
      'int': 2,
      'list': [1, 2],
      'set': {1, 2},
      'tuple': (1, 2),
      'ref': config_dict.FieldReference({'int': 0}),
      'inner_dict_1': inner_dict,
      'inner_dict_2': inner_dict
  }

  print_section('Initializing on dictionary.')
  # ConfigDict can be initialized on example_dict
  example_cd = config_dict.ConfigDict(example_dict)

  # Dictionary fields are also converted to ConfigDict
  print(type(example_cd.inner_dict_1))

  # And the reference structure is preserved
  print(id(example_cd.inner_dict_1) == id(example_cd.inner_dict_2))

  print_section('Initializing on ConfigDict.')

  # ConfigDict can also be initialized on a ConfigDict
  example_cd_cd = config_dict.ConfigDict(example_cd)

  # Yielding the same result:
  print(example_cd == example_cd_cd)

  # Note that the memory addresses are different
  print(id(example_cd) == id(example_cd_cd))

  # The memory addresses of the attributes are not the same because of the
  # FieldReference, which gets removed on the second initialization
  list_to_ids = lambda x: [id(element) for element in x]
  print(
      set(list_to_ids(list(example_cd.values()))) == set(
          list_to_ids(list(example_cd_cd.values()))))

  print_section('Initializing on self-referencing dictionary.')

  # Initialization works on a self-referencing dict
  self_ref_dict = copy.deepcopy(example_dict)
  self_ref_dict['self'] = self_ref_dict
  self_ref_cd = config_dict.ConfigDict(self_ref_dict)

  # And the reference structure is replicated
  print(id(self_ref_cd) == id(self_ref_cd.self))

  print_section('Unexpected initialization behavior.')

  # ConfigDict initialization doesn't look inside lists, so doesn't convert a
  # dict in a list to ConfigDict
  dict_in_list_in_dict = {'list': [{'troublemaker': 0}]}
  dict_in_list_in_dict_cd = config_dict.ConfigDict(dict_in_list_in_dict)
  print(type(dict_in_list_in_dict_cd.list[0]))

  # This can cause the reference structure to not be replicated
  referred_dict = {'key': 'value'}
  bad_reference = {'referred_dict': referred_dict, 'list': [referred_dict]}
  bad_reference_cd = config_dict.ConfigDict(bad_reference)
  print(id(bad_reference_cd.referred_dict) == id(bad_reference_cd.list[0]))


if __name__ == '__main__':
  app.run()



# Copyright 2024 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A enum to be used in a mock_config.

The enum can't be defined directly in the config because a config imported
through a flag is a separate instance of the config module. Therefore it would
define an own instance of the enum class which won't be equal to the same enum
from the config imported directly.
"""

import enum


class SporkType(enum.Enum):
  SPOON = 1
  SPORK = 2
  FORK = 3



# Copyright 2024 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Configuration commmand line parser."""

import ast
import copy
import dataclasses
import enum
import errno
import functools as ft
import importlib.machinery
import os
import re
import sys
import traceback
import types
from typing import Any, Callable, Dict, Generic, List, MutableMapping, Optional, Sequence, Tuple, Type, TypeVar

from absl import flags
from absl import logging
from ml_collections import config_dict
from ml_collections.config_flags import config_path
from ml_collections.config_flags import tuple_parser

FLAGS = flags.FLAGS

# Forward for backwards compatibility.
GetValue = config_path.get_value
GetType = config_path.get_type
SetValue = config_path.set_value

# Any flags defined via any of the config_flags.DEFINE_config_*() functions
# should be attributed to the caller of that function rather than this module.
flags.disclaim_key_flags()


def _load_source(module_name: str, module_path: str) -> types.ModuleType:
  """Loads a Python module from its source file.

  Args:
    module_name: name of the module in sys.modules.
    module_path: path to the Python file containing the module.

  Returns:
    The loaded Python module.
  """
  loader = importlib.machinery.SourceFileLoader(module_name, module_path)
  return loader.load_module()


class _LiteralParser(flags.ArgumentParser):
  """Parse arbitrary built-in (`--cfg.val=1`, `--cfg.val="[1, 2, {}]"`,...)."""

  def parse(self, argument: str) -> Any:
    # _LiteralParser cannot know in advance what is the expected type.
    # The default value is never passed, as default is overwritten to `None`
    # bellow inside `_ConfigFlag._parse`.
    if not isinstance(argument, str):
      raise TypeError('argument should be a string')
    # Absl hardcode bool values as lower-case: `--cfg.my_bool`, so convert
    # them to Python built-in
    if argument in ('true', 'false'):
      argument = argument.capitalize()
    try:
      return ast.literal_eval(argument)
    except (SyntaxError, ValueError):
      # Otherwise, the flag is a string: `--cfg.value="my_string"`
      return argument

  def flag_type(self):
    return 'config_literal'


class _ConfigDictParser(flags.ArgumentParser):
  """Parser for ConfigDict values."""

  def parse(self, argument: str) -> config_dict.ConfigDict:
    try:
      value = ast.literal_eval(argument)
    except (SyntaxError, ValueError) as e:
      raise ValueError(
          f'Failed to parse {argument!r} as a ConfigDict: {e!r}'
      ) from None

    if not isinstance(value, dict):
      raise ValueError(
          f'Failed to parse {argument!r} as a ConfigDict: `{value!r}` is not a'
          ' dict.'
      )
    return config_dict.ConfigDict(value)


_FIELD_TYPE_TO_PARSER = {
    float: flags.FloatParser(),
    bool: flags.BooleanParser(),
    tuple: tuple_parser.TupleParser(),
    int: flags.IntegerParser(),
    str: flags.ArgumentParser(),
    config_dict.ConfigDict: _ConfigDictParser(),
    object: _LiteralParser(),
}


class UnsupportedOperationError(flags.Error):
  pass


class FlagOrderError(flags.Error):
  pass


class UnparsedFlagError(flags.Error):
  pass


def DEFINE_config_file(  # pylint: disable=g-bad-name
    name: str,
    default: Optional[str] = None,
    help_string: str = 'path to config file.',
    flag_values: flags.FlagValues = FLAGS,
    lock_config: bool = True,
    accept_new_attributes: bool = False,
    sys_argv: Optional[List[str]] = None,
    **kwargs) -> flags.FlagHolder:
  r"""Defines flag for `ConfigDict` files compatible with absl flags.

  The flag's value should be a path to a valid python file which contains a
  function called `get_config()` that returns a python object specifying
  a configuration. After the flag is parsed, `FLAGS.name` will contain
  a reference to this object, optionally with some values overridden.

  During flags parsing, every flag of form `--name.([a-zA-Z0-9]+\.?)+=value`
  and `-name.([a-zA-Z0-9]+\.?)+ value` will be treated as an override of a
  specific field in the config object returned by this flag. Field is
  essentially a dot delimited path inside the object where each path element
  has to be either an attribute or a key existing in the config object.
  For example `--my_config.field1.field2=val` means "assign value val
  to the attribute (or key) `field2` inside value of the attribute (or key)
  `field1` inside the value of `my_config` object". If there are both
  attribute and key-based access with the same name, attribute is preferred.

  Typical usage example:

  `script.py`::

    from ml_collections import config_flags

    _CONFIG = config_flags.DEFINE_config_file('my_config')

    print(_CONFIG.value)

  `config.py`::

    def get_config():
      return {
          'field1': 1,
          'field2': 'tom',
          'nested': {
              'field': 2.23,
          },
      }

  The following command::

    python script.py -- --my_config=config.py
                        --my_config.field1 8
                        --my_config.nested.field=2.1

  will print::

    {'field1': 8, 'field2': 'tom', 'nested': {'field': 2.1}}

  It is possible to parameterise the get_config function, allowing it to
  return a differently structured result for different occasions. This is
  particularly useful when setting up hyperparameter sweeps across various
  network architectures.

  `parameterised_config.py`::

    def get_config(config_string):
      possible_configs = {
          'mlp': {
              'constructor': 'snt.nets.MLP',
              'config': {
                  'output_sizes': (128, 128, 1),
              }
          },
          'lstm': {
              'constructor': 'snt.LSTM',
              'config': {
                  'hidden_size': 128,
                  'forget_bias': 1.0,
              }
          }
      }
      return possible_configs[config_string]

  If a colon is present in the command line override for the config file,
  everything to the right of the colon is passed into the get_config function.
  The following command lines will both function correctly::

    python script.py -- --my_config=parameterised_config.py:mlp
                        --my_config.config.output_sizes="(256,256,1)"


    python script.py -- --my_config=parameterised_config.py:lstm
                        --my_config.config.hidden_size=256

  The following will produce an error, as the hidden_size flag does not
  exist when the "mlp" config_string is provided::

    python script.py -- --my_config=parameterised_config.py:mlp
                        --my_config.config.hidden_size=256

  Args:
    name: Flag name, optionally including extra config after a colon.
    default: Default value of the flag (default: None).
    help_string: Help string to display when --helpfull is called. (default:
      "path to config file.")
    flag_values: FlagValues instance used for parsing. (default:
      absl.flags.FLAGS)
    lock_config: If set to True, loaded config will be locked through calling
      .lock() method on its instance (if it exists). (default: True)
    accept_new_attributes: If `True`, accept to pass arbitrary attributes that
      are not originally defined in the `get_config()` dict.
      `accept_new_attributes` requires `lock_config=False`.
    sys_argv: If set, interprets this as the full list of args used in parsing.
      This is used to identify which overrides to define as flags. If not
      specified, uses the system sys.argv to figure it out.
    **kwargs: Optional keyword arguments passed to Flag constructor.

  Returns:
    a handle to defined flag.
  """
  if accept_new_attributes and lock_config:
    raise ValueError('`accept_new_attributes=True` requires lock_config=False')
  parser = ConfigFileFlagParser(name=name, lock_config=lock_config)
  serializer = flags.ArgumentSerializer()
  flag = _ConfigFlag(
      parser=parser,
      serializer=serializer,
      name=name,
      default=default,
      help_string=help_string,
      flag_values=flag_values,
      accept_new_attributes=accept_new_attributes,
      sys_argv=sys_argv,
      **kwargs)

  return flags.DEFINE_flag(flag, flag_values)


def DEFINE_config_dict(  # pylint: disable=g-bad-name
    name: str,
    config: config_dict.ConfigDict,
    help_string: str = 'ConfigDict instance.',
    flag_values: flags.FlagValues = FLAGS,
    lock_config: bool = True,
    accept_new_attributes: bool = False,
    sys_argv: Optional[List[str]] = None,
    **kwargs) -> flags.FlagHolder:
  """Defines flag for inline `ConfigDict's` compatible with absl flags.

  Similar to `DEFINE_config_file` except the flag's value should be a
  `ConfigDict` instead of a path to a file containing a `ConfigDict`. After the
  flag is parsed, `FLAGS.name` will contain a reference to the `ConfigDict`,
  optionally with some values overridden.

  Typical usage example:

  `script.py`::

    from ml_collections import config_dict
    from ml_collections import config_flags


    config = config_dict.ConfigDict({
        'field1': 1,
        'field2': 'tom',
        'nested': {
            'field': 2.23,
        }
    })


    _CONFIG = config_flags.DEFINE_config_dict('my_config', config)
    ...

    print(_CONFIG.value)

  The following command::

    python script.py -- --my_config.field1 8
                        --my_config.nested.field=2.1

  will print::

    field1: 8
    field2: tom
    nested: {field: 2.1}

  Args:
    name: Flag name.
    config: `ConfigDict` object.
    help_string: Help string to display when --helpfull is called.
        (default: "ConfigDict instance.")
    flag_values: FlagValues instance used for parsing.
        (default: absl.flags.FLAGS)
    lock_config: If set to True, loaded config will be locked through calling
        .lock() method on its instance (if it exists). (default: True)
    accept_new_attributes: If `True`, accept to pass arbitrary attributes that
      are not originally defined in the `config` argument.
      `accept_new_attributes` requires `lock_config=False`.
    sys_argv: If set, interprets this as the full list of args used in parsing.
      This is used to identify which overrides to define as flags. If not
      specified, uses the system sys.argv to figure it out.
    **kwargs: Optional keyword arguments passed to Flag constructor.

  Returns:
    a handle to defined flag.
  """
  if not isinstance(config, config_dict.ConfigDict):
    raise TypeError('config should be a ConfigDict')
  if accept_new_attributes and lock_config:
    raise ValueError('`accept_new_attributes=True` requires lock_config=False')
  parser = _InlineConfigParser(name=name, lock_config=lock_config)
  flag = _ConfigFlag(
      parser=parser,
      serializer=flags.ArgumentSerializer(),
      name=name,
      default=config,
      help_string=help_string,
      flag_values=flag_values,
      accept_new_attributes=accept_new_attributes,
      sys_argv=sys_argv,
      **kwargs)

  # Get the module name for the frame at depth 1 in the call stack.
  module_name = sys._getframe(1).f_globals.get('__name__', None)  # pylint: disable=protected-access
  module_name = sys.argv[0] if module_name == '__main__' else module_name
  return flags.DEFINE_flag(flag, flag_values, module_name=module_name)


# Note that we would add a bound to constrain this to be a dataclass, except
# that dataclasses don't have a specific base class, and structural typing for
# attributes is currently (2021Q1) not supported in pytype (b/150927776).
_T = TypeVar('_T')


class  _DataclassParser(flags.ArgumentParser, Generic[_T]):
  """Parser for a config defined inline (not from a file)."""

  def __init__(self, name: str, dataclass_type: Type[_T],
               parse_fn: Optional[Callable[[Any], _T]] = None):
    self.name = name
    self.dataclass_type = dataclass_type
    self.parse_fn = parse_fn

  def _get_parse_fn(self):
    # NB: We are using lazy lookup in the parse function rather than
    # constructor, to give client's code a chance to create their parsers
    # and provide consistency with availability of parsers for nested fields
    # (which are also looked up during parsing).
    if self.parse_fn:
      return self.parse_fn
    if self.dataclass_type in _FIELD_TYPE_TO_PARSER:
      return _FIELD_TYPE_TO_PARSER[self.dataclass_type].parse

  def parse(self, config: Any) -> _T:
    # It is important to use deepcopy here, so if parser returns constants
    # they are not modified during the flag parsing.
    if isinstance(config, self.dataclass_type):
      return copy.deepcopy(config)
    parse_fn = self._get_parse_fn()
    if parse_fn:
      return copy.deepcopy(parse_fn(config))
    raise TypeError(
        'Overriding {} is not allowed: it has neither '
        'registered parser nor explicit parse_fn in the flag '
        'definition'.format(self.name)
    )

  def flag_type(self):
    return 'config_dataclass({})'.format(self.dataclass_type)


def DEFINE_config_dataclass(  # pylint: disable=invalid-name
    name: str,
    config: _T,
    help_string: str = 'Configuration object. Must be a dataclass.',
    flag_values: flags.FlagValues = FLAGS,
    sys_argv: Optional[List[str]] = None,
    parse_fn: Optional[Callable[[Any], _T]] = None,
    **kwargs,
) -> flags.FlagHolder[_T]:
  """Defines a typed (dataclass) flag-overrideable configuration.

  Similar to `DEFINE_config_dict` except `config` should be a `dataclass`.

  The config value can contain nested fields, including other dataclasses.
  If a field is of form  Optional[dataclass] with None as a default value,
  it can be explicitly initialized using special value `build`. E.g.
  For instance:

  ```
  @dc.dataclass
  class FancyLoss
    foo_scale: float = 0.1

  @dc.dataclass
  class Config:
    fancy_loss: Optional[FancyLoss] = None
  ```

  Then if `--config.fancy_loss=build  --config.fancy_loss.foo_scale=1` will
  instantiate and override foo_scale to 1.  Note: that the reverse
  order is not allowed:  `--config.fancy_loss.foo_scale=1
  --config.fancy_loss=build` will cause FlagOrderError.

  Optional dataclass fields can also be set to None using special `none` value.
  For instance:

  ```
  @dc.dataclass
  class FancyLossConfig
    foo_loss_scale: float = 0.1

  @dc.dataclass
  class Config:
    fancy_loss: Optional[FancyLossConfig] = FancyLossConfig()
  ```

  Then `--config.fancy_loss=none`, will set it to None.

  Implementation note: This flag will register all the needed nested flags
  dynamically based on sys.argv or sys_argv, in order to support
  free-text keyed flags such as `foo.bar['i']=1`. Because of how flag subsystem
  works this will happen either at flag-parsing time (during app.run),
  if there is a root level override, such as `--config=<..>` or during
  declaration otherwise (during this invocation).
  Parsing at declaration (e.g. if no root override) can cause problems
  with multiprocessing  since sys.argv is not yet populated at
  declaration time for spawn processes. To avoid this, pass custom sys_argv
  value instead if you want to use this library with multiprocessing.
  Also note, in the future we might consider to always do it at declaration
  time, as this cleans up the logic significantly.

  Args:
    name: Flag name.
    config: A user-defined configuration object. Must be built via `dataclass`.
    help_string: Help string to display when --helpfull is called.
    flag_values: FlagValues instance used for parsing.
    sys_argv: If set, interprets this as the full list of args used in parsing.
      This is used to identify which overrides to define as flags. If not
      specified, uses the system sys.argv to figure it out.
    parse_fn: Function that can parse provided flag value, when assigned via
      flag.value, or passed on command line. If not provided, but the class has
      registered parser register_flag_parser_for_type, the latter will be used.
      Otherwise only allows to assign instances of this class.
    **kwargs: Optional keyword arguments passed to Flag constructor.

  Returns:
    A handle to the defined flag.
  """

  if not dataclasses.is_dataclass(config):
    raise ValueError('Configuration object must be a `dataclass`.')
  # Define the flag.
  parser = _DataclassParser(name=name, dataclass_type=type(config),
                            parse_fn=parse_fn)
  flag = _ConfigFlag(
      flag_values=flag_values,
      parser=parser,
      serializer=flags.ArgumentSerializer(),
      name=name,
      default=config,
      help_string=help_string,
      sys_argv=sys_argv,
      **kwargs)

  return flags.DEFINE_flag(flag, flag_values)


def get_config_filename(config_flag) -> str:  # pylint: disable=g-bad-name
  """Returns the path to the config file given the config flag.

  Args:
    config_flag: The flag instance obtained from FLAGS, e.g. FLAGS['config'].

  Returns:
    the path to the config file.
  """
  if not is_config_flag(config_flag):
    raise TypeError('expect a config flag, found {}'.format(type(config_flag)))
  return config_flag.config_filename


def get_override_values(config_flag) -> Dict[str, Any]:  # pylint: disable=g-bad-name
  """Returns a flat dict containing overridden values from the config flag.

  Args:
    config_flag: The flag instance obtained from FLAGS, e.g. FLAGS['config'].

  Returns:
    a flat dict containing overridden values from the config flag.
  """
  if not is_config_flag(config_flag):
    raise TypeError('expect a config flag, found {}'.format(type(config_flag)))
  return config_flag.override_values


class _IgnoreFileNotFoundAndCollectErrors:
  """Helps recording "file not found" exceptions when loading config.

  Usage:
    ignore_errors = _IgnoreFileNotFoundAndCollectErrors()
    with ignore_errors.Attempt('Loading from foo', 'bar.id'):
      ...
      return True  # successfully loaded from `foo`
    logging.error('Failed loading: {}'.format(ignore_errors.DescribeAttempts()))
  """

  def __init__(self):
    self._attempts = []  # type: List[Tuple[Tuple[str, str], IOError]]

  def Attempt(self, description, path):
    """Creates a context manager that routes exceptions to this class."""
    self._current_attempt = (description, path)
    ignore_errors = self

    class _ContextManager:

      def __enter__(self):
        return self

      def __exit__(self, exc_type, exc_value, unused_traceback):
        return ignore_errors.ProcessAttemptException(exc_type, exc_value)

    return _ContextManager()

  def ProcessAttemptException(self, exc_type, exc_value):
    if exc_type is FileNotFoundError and exc_value.errno == errno.ENOENT:  # pytype: disable=attribute-error  # trace-all-classes
      self._attempts.append((self._current_attempt, exc_value))  # pytype: disable=container-type-mismatch  # trace-all-classes
      # Returning a true value suppresses exceptions:
      # https://docs.python.org/2/reference/datamodel.html#object.__exit__
      return True

  def DescribeAttempts(self):
    return '\n'.join(
        '  Attempted [{}]:\n    {}\n      {}'.format(attempt[0], attempt[1], e)
        for attempt, e in self._attempts)


def _MakeDefaultOrNone(kls, config, allow_none=True, field_path=''):
  if config in ['build', True]:
    try:
      return kls()
    except Exception as e:
      raise ValueError(
          f'Unable to create default instance for "{field_path}" '
          f'of type "{kls}": {e}') from e

  elif (config in ['0', 0, False] or config.lower() == 'none'):
    if not allow_none:
      raise ValueError(f'None is not allowed as value for "{field_path}", '
                       'as the dataclass field is not marked as optional.')
    return None
  raise ValueError(f'Unable to parse value "{config}" as instance of {kls}'
                   f'for {field_path} values allowed are [0/none, or 1]')


def _LoadConfigModule(name: str, path: str):
  """Loads a script from external file specified by path.

  Unprefixed path is looked for in the current working directory using
  regular file open operation. This should work with relative config paths.

  Args:
    name: Name of the new module.
    path: Path to the .py file containing the module.

  Returns:
    Module loaded from the given path.

  Raises:
    IOError: If the config file cannot be found.
  """
  if not path:
    raise IOError('Path to config file is an empty string.')

  ignoring_errors = _IgnoreFileNotFoundAndCollectErrors()

  # Works for relative paths.
  with ignoring_errors.Attempt('Relative path', path):
    config_module = _load_source(name, path)
    return config_module

  # Nothing worked. Log the paths that were attempted.
  raise IOError('Failed loading config file {}\n{}'.format(
      name, ignoring_errors.DescribeAttempts()))


class _ErrorConfig:
  """ConfigDict object that raises an error on any attribute access."""

  def __init__(self, error):
    super(_ErrorConfig, self).__init__()
    super(_ErrorConfig, self).__setattr__('_error', error)

  def __getattr__(self, attr):
    self._ReportError()

  def __setattr__(self, attr, value):
    self._ReportError()

  def __delattr__(self, attr):
    self._ReportError()

  def __getitem__(self, key):
    self._ReportError()

  def __setitem__(self, key, value):
    self._ReportError()

  def __delitem__(self, key):
    self._ReportError()

  def _ReportError(self):  # pylint: disable=invalid-name
    raise IOError(
        'Configuration is not available because of an earlier failure to load:'
        f' {self._error}'
    ) from self._error


def _LockConfig(config):
  """Calls config.lock() if config has a lock method."""
  if isinstance(config, _ErrorConfig):
    pass  # Attempting to access _ErrorConfig.lock will raise its error.
  elif getattr(config, 'lock', None) and callable(config.lock):
    config.lock()
  else:
    pass  # config.lock() does not have desired semantics, do nothing.


class ConfigFileFlagParser(flags.ArgumentParser):
  """Parser for config files."""

  def __init__(self, name, lock_config=True):
    self.name = name
    self._lock_config = lock_config

  def parse(self, path):
    """Loads a config module from `path` and returns the `get_config()` result.

    If a colon is present in `path`, everything to the right of the first colon
    is passed to `get_config` as an argument. This allows the structure of what
    is returned to be modified, which is useful when performing complex
    hyperparameter sweeps.

    Args:
      path: string, path pointing to the config file to execute. May also
          contain a config_string argument, e.g. be of the form
          "config.py:some_configuration".
    Returns:
      Result of calling `get_config` in the specified module.
    """
    # This will be a 2 element list iff extra configuration args are present.
    split_path = path.split(':', 1)
    try:
      config_module = _LoadConfigModule('{}_config'.format(self.name),
                                        split_path[0])
      config = config_module.get_config(*split_path[1:])
      if config is None:
        logging.warning(
            '%s:get_config() returned None, did you forget a return statement?',
            path)
    except IOError as e:
      # Don't raise the error unless/until the config is actually accessed.
      config = _ErrorConfig(e)
    # Third party flags library catches TypeError and ValueError and rethrows,
    # removing useful information unless it is added here (b/63877430):
    except (TypeError, ValueError) as e:
      error_trace = traceback.format_exc()
      raise type(e)('Error whilst parsing config file:\n\n' + error_trace)

    if self._lock_config:
      _LockConfig(config)

    return config

  def flag_type(self):
    return 'config object'


# Alias to an older name, for backwards compatibility.
_ConfigFileParser = ConfigFileFlagParser


class _InlineConfigParser(flags.ArgumentParser):
  """Parser for a config defined inline (not from a file)."""

  def __init__(self, name, lock_config=True):
    self.name = name
    self._lock_config = lock_config

  def parse(self, config):
    if not isinstance(config, config_dict.ConfigDict):
      raise TypeError('Overriding {} is not allowed.'.format(self.name))
    if self._lock_config:
      _LockConfig(config)
    return config

  def flag_type(self):
    return 'config object'


class _ConfigFlag(flags.Flag):
  """Flag definition for command-line overridable configs."""

  def __init__(
      self,
      flag_values=FLAGS,
      *,
      accept_new_attributes: bool = False,
      sys_argv=None,
      **kwargs,
  ):
    # Parent constructor can already call .Parse, thus additional fields
    # have to be set here.
    self.flag_values = flag_values
    self._accept_new_attributes = accept_new_attributes
    # Note, we don't replace sys_argv with sys.argv here if it's None because
    # in some obscure multiprocessing use cases, sys.argv may not be populated
    # until later and we need to look it up at parse time.
    self._sys_argv = sys_argv
    super(_ConfigFlag, self).__init__(**kwargs)

  def _GetArgv(self):
    """Lazily fetches sys.argv and expands any potential --flagfile=..."""
    argv = sys.argv if self._sys_argv is None else self._sys_argv
    argv = flags.FLAGS.read_flags_from_files(argv, force_gnu=False)
    return argv

  def _GetOverrides(self, argv):
    """Parses the command line arguments for the overrides."""
    # We use a dict to keep the order of the overrides.
    overrides = dict()
    config_index = self._FindConfigSpecified(argv)
    for i, arg in enumerate(argv):
      if re.match(r'-{{1,2}}(no)?{}\.'.format(self.name), arg):
        if config_index > 0 and i < config_index:
          raise FlagOrderError('Found {} in argv before a value for --{} '
                               'was specified'.format(arg, self.name))
        arg_name = arg.split('=', 1)[0]
        override_name = arg_name.split('.', 1)[1]
        # Reinsert override_name into overrides to keep the order of
        # the "last" flag.
        overrides[override_name] = overrides.pop(override_name, None)
    return list(overrides)

  def _FindConfigSpecified(self, argv):
    """Finds element in argv specifying the value of the config flag.

    Args:
      argv: command line arguments as a list of strings.
    Returns:
      Index in argv if found and -1 otherwise.
    """
    for i, arg in enumerate(argv):
      # '-(-)config' followed by '=' or at the end of the string.
      if re.match(r'^-{{1,2}}{}(=|$)'.format(self.name), arg) is not None:
        return i
    return -1

  def _IsConfigSpecified(self, argv):
    """Returns `True` if the config file is specified on the command line."""
    return self._FindConfigSpecified(argv) >= 0

  def _set_default(self, default):
    if self._IsConfigSpecified(self._GetArgv()):
      self.default = default
    else:
      super(_ConfigFlag, self)._set_default(default)  # pytype: disable=attribute-error
    self.default_as_str = "'{}'".format(default)

  def _validate_overrides(self, config, overrides: Sequence[str]):
    # Verify that we don't provide --config.foo.bar=1 followed by override of
    # config.foo.
    for i, override_a in enumerate(overrides, 1):
      for override_b in overrides[i:]:
        # verify if override_b will overwrite override_a
        if override_a.startswith(override_b + '.'):
          raise FlagOrderError(
              f'Flag --{self.name}.{override_b} is provided after '
              f'--{self.name}.{override_a} and '
              'it will overwrite the value provided in '
              f'--{self.name}.{override_a}, '
              'which is probably not what you expect.')

  def _initialize_missing_parent_fields(self, config, overrides):
    for override in overrides:
      config_path.initialize_missing_parent_fields(
          config, override, overrides)

  def _parse(self, argument):
    # Parse config
    config = super(_ConfigFlag, self)._parse(argument)

    # Get list or overrides
    overrides = self._GetOverrides(self._GetArgv())
    # Iterate over overridden fields and create valid parsers
    self._override_values = {}
    self._initialize_missing_parent_fields(config, overrides)
    self._validate_overrides(config, overrides)

    if self._accept_new_attributes:
      # If user provide a new attribute, fallback to `object` to accept all
      # literal
      default_type = object
    else:
      default_type = None

    for field_path in overrides:
      field_type = config_path.get_type(
          field_path, config, default_type=default_type
      )
      field_type_origin = config_path.get_origin(field_type)
      field_help = 'An override of {}\'s field {}'.format(self.name, field_path)
      field_name = '{}.{}'.format(self.name, field_path)

      parser = None
      if field_type in _FIELD_TYPE_TO_PARSER:
        parser = _FIELD_TYPE_TO_PARSER[field_type]
      elif isinstance(field_type, type) and issubclass(
          field_type, config_dict.ConfigDict
      ):
        # Supports ConfigDict sub-classes.
        parser = _FIELD_TYPE_TO_PARSER[config_dict.ConfigDict]
      elif field_type_origin and field_type_origin in _FIELD_TYPE_TO_PARSER:
        parser = _FIELD_TYPE_TO_PARSER[field_type_origin]
      elif issubclass(field_type, enum.Enum):
        parser = flags.EnumClassParser(field_type, case_sensitive=False)
      elif dataclasses.is_dataclass(field_type):
        # For dataclasses-valued fields allow default instance creation.
        is_optional = config_path.is_optional(field_path, config)
        parser = _DataclassParser(
            name=field_path, dataclass_type=field_type,
            parse_fn=ft.partial(_MakeDefaultOrNone, field_type,
                                allow_none=is_optional, field_path=field_path))

      if parser:
        if not isinstance(parser, tuple_parser.TupleParser):
          if isinstance(parser, (_LiteralParser, _ConfigDictParser)):
            # We do not pass the default to `_ConfigFieldFlag`, otherwise
            # `_LiteralParser.parse(default)` is called with `default`,
            # which would try to parse string.
            # Setting the value to `None` never call `.parse`, so the
            # default value from the config is kept.
            # TODO(sandler): Investigate if value could be None for all parsers.
            default = None
          else:
            default = config_path.get_value(field_path, config)
          flag = _ConfigFieldFlag(
              path=field_path,
              config=config,
              override_values=self._override_values,
              parser=parser,
              serializer=flags.ArgumentSerializer(),
              name=field_name,
              default=default,
              accept_new_attributes=self._accept_new_attributes,
              help_string=field_help,
          )
          # Literal values support the `--my_bool` / `--nomy_bool` syntax
          flag.boolean = field_type is bool or isinstance(
              parser, _LiteralParser
          )
          flags.DEFINE_flag(flag=flag, flag_values=self.flag_values)
        elif field_name not in self.flag_values:
          # Overriding a tuple field. Define the flag only once -- it might
          # appear multiple times on the command-line (e.g.
          # `--config.flag a --config.flag b`) but defining the same flag
          # multiple times is an error. All arguments for the same flag are
          # passed to a single call of _ConfigFieldMultiFlag.parse.
          flag = _ConfigFieldMultiFlag(
              path=field_path,
              config=config,
              override_values=self._override_values,
              parser=parser,
              serializer=flags.ArgumentSerializer(),
              name=field_name,
              default=config_path.get_value(field_path, config),
              help_string=field_help,
          )
          flags.DEFINE_flag(flag=flag, flag_values=self.flag_values)
      else:
        raise UnsupportedOperationError(
            "Type {} of field {} is not supported for overriding. "
            "Currently supported types are: {}. (Note that tuples should "
            "be passed as a string on the command line, `--flag='(a, b, c)'`, "
            "or by repeated flags, `--flag=1 --flag=2 --flag=3`, rather than "
            "--flag=(a, b, c).)".format(
                field_type, field_name, _FIELD_TYPE_TO_PARSER.keys()))

    self._config_filename = argument
    return config

  def serialize(self):
    # Use the config filename instead of the dictionary when serializing.
    return self._serialize(self.config_filename)

  @property
  def config_filename(self):
    """Returns a path to a config file.

    Typical usage example:
    `script.py`:

    ```python
    ...
    from absl import flags
    from ml_collections import config_flags

    FLAGS = flags.FLAGS
    _CONFIG = config_flags.DEFINE_config_file(
      name='my_config',
      default='ml_collections/config_flags/tests/configdict_config.py',
      help_string='config file')
    ...

    FLAGS['my_config'].config_filename

    will output
    'ml_collections/config_flags/tests/configdict_config.py'
    ```

    Returns:
      A path to a config file. For a parameterised get_config, the config
      filename with the provided parameterisation is returned.

    Raises:
      UnparsedFlagError: if the flag has not been parsed.

    """
    if not hasattr(self, '_config_filename'):
      raise UnparsedFlagError('The flag has not been parsed yet')
    return self._config_filename

  @property
  def override_values(self):
    """Returns a flat dictionary containing overridden values.

    Keys in the dictionary are dot-separated paths navigating to child items in
    the original configuration. For example, supppose that a `config` flag is
    defined and initialized to the following configuration:

    ```python
    {
        'a': 1,
        'nested': {
            'b': 2
        }
    }
    ```

    and the user overrides both values using command-line flags:

    ```
    --config.a=10 --config.nested.b=20
    ```

    Then `FLAGS['config'].override_values` will return:

    ```python
    {
        'a': 10,
        'nested.b': 20
    }
    ```

    The result can be passed to `ConfigDict.update_from_flattened_dict` to
    update the values in a configuration. Continuing with the example above:

    ```python
    from ml_collections import config_dict
    config = config_dict.ConfigDict{
        'a': 123,
        'nested': {
            'b': 456
        }
    }
    config.update_from_flattened_dict(FLAGS['config'].override_values)
    print(config.a)  # Prints `10`.
    print(config.nested.b)  # Prints `20`.
    ```

    Returns:
      Flat dictionary with overridden values.

    Raises:
      UnparsedFlagError: if the flag has not been parsed.
    """
    if not hasattr(self, '_override_values'):
      raise UnparsedFlagError('The flag has not been parsed yet')
    return self._override_values


def is_config_flag(flag):  # pylint: disable=g-bad-name
  """Returns True iff `flag` is an instance of `_ConfigFlag`.

  External users of the library may need to check if a flag is of this type
  or not, particularly because ConfigFlags should be parsed before any other
  flags. This function allows that test to be done without making the whole
  class public.

  Args:
    flag: Flag object.

  Returns:
    True iff `isinstance(flag, _ConfigFlag)` is true.
  """
  return isinstance(flag, _ConfigFlag)


class _ConfigFieldFlag(flags.Flag):
  """Flag for updating a field in a ConfigDict."""

  def __init__(
      self,
      path: str,
      config: config_dict.ConfigDict,
      override_values: MutableMapping[str, Any],
      *,
      parser: flags.ArgumentParser,
      serializer: flags.ArgumentSerializer,
      name: str,
      default: Any,
      help_string: str,
      short_name: Optional[str] = None,
      boolean: bool = False,
      accept_new_attributes: bool = False,
  ):
    """Creates new flag with callback."""
    super().__init__(
        parser=parser,
        serializer=serializer,
        name=name,
        default=default,
        help_string=help_string,
        short_name=short_name,
        boolean=boolean)
    self._path = path
    self._config = config
    self._override_values = override_values
    self._accept_new_attributes = accept_new_attributes

  def parse(self, argument):
    super().parse(argument)
    # Callback to set value in ConfigDict.
    config_path.set_value(
        self._path, self._config, self.value,
        accept_new_attributes=self._accept_new_attributes,
    )
    self._override_values[self._path] = self.value


class _ConfigFieldMultiFlag(flags.MultiFlag):
  """Flag for updating a tuple field in a ConfigDict."""

  def __init__(
      self,
      path: str,
      config: config_dict.ConfigDict,
      override_values: MutableMapping[str, Any],
      *,
      parser: flags.ArgumentParser,
      serializer: flags.ArgumentSerializer,
      name: str,
      default: Any,
      help_string: str,
      short_name: Optional[str] = None,
      boolean: bool = False,
  ):
    """Creates new flag with callback."""
    super().__init__(
        parser=parser,
        serializer=serializer,
        name=name,
        default=default,
        help_string=help_string,
        short_name=short_name,
        boolean=boolean)
    self._path = path
    self._config = config
    self._override_values = override_values

  def parse(self, arguments):
    super().parse(arguments)
    # Callback to set value in ConfigDict.
    config_path.set_value(self._path, self._config, tuple(self.value))
    self._override_values[self._path] = tuple(self.value)

  def _parse(self, arguments):
    # MultiFlag passes each argument one-at-a-time to the parser.parse. Just
    # call Flag._parse (grandparent class) directly so all arguments are passed
    # to parser.parse in a single call.
    result = flags.Flag._parse(self, arguments)  # pylint: disable=protected-access
    return list(result)


def register_flag_parser_for_type(
    field_type: _T, parser: flags.ArgumentParser) -> _T:
  """Registers parser for a given type.

  See documentation for `register_flag_parser` for usage example.

  Args:
    field_type: field type to register
    parser: parser to use

  Returns:
    field_type unmodified.
  """
  _FIELD_TYPE_TO_PARSER[field_type] = parser
  return field_type


def register_flag_parser(*, parser: flags.ArgumentParser) -> Callable[[_T], _T]:
  """Creates a decorator to register parser on types.

  For example:

  ```
  class ParserForCustomConfig(flags.ArgumentParser):
  def parse(self, value):
    if isinstance(value, CustomConfig):
      return value
    return CustomConfig(i=int(value), j=int(value))


  @config_flags.register_flag_parser(parser=ParserForCustomConfig())
  @dataclasses.dataclass
  class CustomConfig:
    i: int = None
    j: int = None

  class MainConfig:
    sub: CustomConfig = CustomConfig()

  config_flags.DEFINE_config_dataclass(
    'cfg', MainConfig(), 'MyConfig data')
  ```

  will declare cfg flag, then passing `--cfg.sub=1`, will initialize
  both i and j fields to 1. The fields can still be set individually:
  `--cfg.sub=1 --cfg.sub.j=3` will set `i` to `1` and `j` to `3`.

  Args:
    parser: parser to use.

  Returns:
    Decorator to apply to types.
  """
  return ft.partial(register_flag_parser_for_type, parser=parser)



# Copyright 2024 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for config_flags used in conjunction with DEFINE_config_dataclass."""

import copy
import dataclasses
import functools
import sys
from typing import Mapping, Optional, Sequence, Tuple, Union
import unittest

from absl import flags
from absl.testing import absltest
from ml_collections import config_flags
from ml_collections.config_flags import config_flags as config_flag_lib


#####
# Simple configuration class for testing.
@dataclasses.dataclass
class MyModelConfig:
  foo: int
  bar: Sequence[str]
  baz: Optional[Mapping[str, str]] = None
  buz: Optional[Mapping[Tuple[int, int], str]] = None
  qux: Optional[int] = None
  bax: float = 1
  boj: Tuple[int, ...] = ()


class ParserForCustomConfig(flags.ArgumentParser):
  def __init__(self, delta=1):
    self.delta = delta

  def parse(self, value):
    if isinstance(value, CustomParserConfig):
      return value
    return CustomParserConfig(i=int(value), j=int(value) + self.delta)


@dataclasses.dataclass
@config_flags.register_flag_parser(parser=ParserForCustomConfig())
class CustomParserConfig():
  i: int
  j: int = 1


@dataclasses.dataclass
class MyConfig:
  my_model: MyModelConfig
  baseline_model: MyModelConfig
  custom: CustomParserConfig = dataclasses.field(
      default_factory=lambda: CustomParserConfig(0))


@dataclasses.dataclass
class SubConfig:
  model: Optional[MyModelConfig] = dataclasses.field(
      default_factory=lambda: MyModelConfig(foo=0, bar=['1']))


@dataclasses.dataclass
class ConfigWithOptionalNestedField:
  sub: Optional[SubConfig] = None
  non_optional: SubConfig = dataclasses.field(
      default_factory=SubConfig)

_CONFIG = MyConfig(
    my_model=MyModelConfig(
        foo=3,
        bar=['a', 'b'],
        baz={'foo.b': 'bar'},
        buz={(0, 0): 'ZeroZero', (0, 1): 'ZeroOne'}
    ),
    baseline_model=MyModelConfig(
        foo=55,
        bar=['c', 'd'],
    ),
)

_TEST_FLAG = config_flags.DEFINE_config_dataclass('test_flag', _CONFIG,
                                                  'MyConfig data')


def _test_flags(default, *flag_args, parse_fn=None):
  flag_values = flags.FlagValues()
  # DEFINE_config_dataclass accesses sys.argv to build flag list!
  old_args = list(sys.argv)
  sys.argv[:] = ['', *['--test_config' + f for f in flag_args]]
  try:
    result = config_flags.DEFINE_config_dataclass(
        'test_config', default, flag_values=flag_values, parse_fn=parse_fn)
    _, *remaining = flag_values(sys.argv)
    if remaining:
      raise ValueError(f'{remaining}')
    # assert not remaining
    return result.value
  finally:
    sys.argv[:] = old_args


def parse_config_flag(value):
  cfg = _CONFIG
  return dataclasses.replace(
      cfg,
      my_model=dataclasses.replace(cfg.my_model, foo=int(value)))


class TypedConfigFlagsTest(absltest.TestCase):

  def test_types(self):
    self.assertIsInstance(_TEST_FLAG.value, MyConfig)
    self.assertEqual(_TEST_FLAG.value, _CONFIG)
    self.assertIsInstance(flags.FLAGS['test_flag'].value, MyConfig)
    self.assertIsInstance(flags.FLAGS.test_flag, MyConfig)
    self.assertEqual(flags.FLAGS['test_flag'].value, _CONFIG)
    module_name = __name__ if __name__ != '__main__' else sys.argv[0]
    self.assertEqual(
        flags.FLAGS.find_module_defining_flag('test_flag'), module_name)

  def test_instance(self):
    config = _test_flags(_CONFIG)
    self.assertIsInstance(config, MyConfig)
    self.assertEqual(config.my_model, _CONFIG.my_model)
    self.assertEqual(_CONFIG, config)

  def test_flag_config_dataclass_optional(self):
    result = _test_flags(_CONFIG, '.baseline_model.qux=10')
    self.assertEqual(result.baseline_model.qux, 10)
    self.assertIsInstance(result.baseline_model.qux, int)
    self.assertIsNone(result.my_model.qux)

  def test_flag_config_dataclass_repeated_arg_use_last(self):
    result = _test_flags(
        _CONFIG, '.baseline_model.qux=10', '.baseline_model.qux=12'
    )
    self.assertEqual(result.baseline_model.qux, 12)
    self.assertIsInstance(result.baseline_model.qux, int)
    self.assertIsNone(result.my_model.qux)

  def test_custom_flag_parsing_shared_default(self):
    result = _test_flags(_CONFIG, '.baseline_model.foo=324')
    result1 = _test_flags(_CONFIG, '.baseline_model.foo=123')
    # Here we verify that despite using _CONFIG as shared default for
    # result and result1, the final values are not in fact shared.
    self.assertEqual(result.baseline_model.foo, 324)
    self.assertEqual(result1.baseline_model.foo, 123)
    self.assertEqual(_CONFIG.baseline_model.foo, 55)

  def test_custom_flag_parsing_parser_override(self):
    config_flags.register_flag_parser_for_type(
        CustomParserConfig, ParserForCustomConfig(2))
    result = _test_flags(_CONFIG, '.custom=10')
    self.assertEqual(result.custom.i, 10)
    self.assertEqual(result.custom.j, 12)

    # Restore old parser.
    config_flags.register_flag_parser_for_type(
        CustomParserConfig, ParserForCustomConfig())

  @unittest.skipIf(sys.version_info[:2] < (3, 10), 'Need 3.10 to test | syntax')
  def test_pipe_syntax(self):
    @dataclasses.dataclass
    class PipeConfig:
      foo: int | None = None

    result = _test_flags(PipeConfig(), '.foo=32')
    self.assertEqual(result.foo, 32)

  def test_custom_flag_parsing_override_work(self):
    # Overrides still work.
    result = _test_flags(_CONFIG, '.custom.i=10')
    self.assertEqual(result.custom.i, 10)
    self.assertEqual(result.custom.j, 1)

  def test_optional_nested_fields(self):
    with self.assertRaises(ValueError):
      # Implicit creation not allowed.
      _test_flags(ConfigWithOptionalNestedField(), '.sub.model.foo=12')

    # Explicit creation works.
    result = _test_flags(
        ConfigWithOptionalNestedField(), '.sub=build', '.sub.model.foo=12'
    )
    self.assertEqual(result.sub.model.foo, 12)

    # Default initialization support.
    result = _test_flags(ConfigWithOptionalNestedField(), '.sub=build')
    self.assertEqual(result.sub.model.foo, 0)

    # Using default value (None).
    result = _test_flags(ConfigWithOptionalNestedField())
    self.assertIsNone(result.sub)

    with self.assertRaises(config_flag_lib.FlagOrderError):
      # Don't allow accidental overwrites.
      _test_flags(
          ConfigWithOptionalNestedField(), '.sub.model.foo=12', '.sub=build'
      )

  def test_set_to_none_dataclass_fields(self):
    result = _test_flags(
        ConfigWithOptionalNestedField(), '.sub=build', '.sub.model=none'
    )
    self.assertIsNone(result.sub.model, None)

    with self.assertRaises(KeyError):
      # Parent field is set to None (from not None default value),
      # so this is not a valid set of flags.
      _test_flags(
          ConfigWithOptionalNestedField(),
          '.sub=build',
          '.sub.model=none',
          '.sub.model.foo=12',
      )

    with self.assertRaises(KeyError):
      # Parent field is explicitly set to None (with None default value),
      # so this is not a valid set of flags.
      _test_flags(
          ConfigWithOptionalNestedField(), '.sub=none', '.sub.model.foo=12'
      )

  def test_set_none_non_optional_dataclass_fields(self):
    with self.assertRaises(flags.IllegalFlagValueError):
      # Field is not marked as optional so it can't be set to None.
      _test_flags(ConfigWithOptionalNestedField(), '.non_optional=None')

  def test_no_default_initializer(self):
    with self.assertRaises(flags.IllegalFlagValueError):
      _test_flags(ConfigWithOptionalNestedField(), '.sub=1', '.sub.model=1')

  def test_custom_flag_parser_invoked(self):
    # custom parser gets invoked
    result = _test_flags(_CONFIG, '.custom=10')
    self.assertEqual(result.custom.i, 10)
    self.assertEqual(result.custom.j, 11)

  def test_custom_flag_parser_invoked_overrides_applied(self):
    result = _test_flags(_CONFIG, '.custom=15', '.custom.i=11')
    # Override applied successfully
    self.assertEqual(result.custom.i, 11)
    self.assertEqual(result.custom.j, 16)

  def test_custom_flag_application_order(self):
    # Disallow for later value to override the earlier value.
    with self.assertRaises(config_flag_lib.FlagOrderError):
      _test_flags(_CONFIG, '.custom.i=11', '.custom=15')

  def test_flag_config_dataclass_type_mismatch(self):
    result = _test_flags(_CONFIG, '.my_model.bax=10')
    self.assertIsInstance(result.my_model.bax, float)
    # We can't do anything when the value isn't overridden.
    self.assertIsInstance(result.baseline_model.bax, int)
    self.assertRaises(
        flags.IllegalFlagValueError,
        functools.partial(_test_flags, _CONFIG, '.my_model.bax=string'),
    )

  def test_illegal_dataclass_field_type(self):

    @dataclasses.dataclass
    class Config:
      field: Union[int, float] = 3

    self.assertRaises(
        TypeError, functools.partial(_test_flags, Config(), '.field=1')
    )

  def test_spurious_dataclass_field(self):

    @dataclasses.dataclass
    class Config:
      field: int = 3
    cfg = Config()
    cfg.extra = 'test'

    self.assertRaises(
        KeyError, functools.partial(_test_flags, cfg, '.extra=hi')
    )

  def test_nested_dataclass(self):

    @dataclasses.dataclass
    class Parent:
      field: int = 3

    @dataclasses.dataclass
    class Child(Parent):
      other: int = 4

    self.assertEqual(_test_flags(Child(), '.field=1').field, 1)

  def test_flag_config_dataclass(self):
    result = _test_flags(_CONFIG, '.baseline_model.foo=10', '.my_model.foo=7')
    self.assertEqual(result.baseline_model.foo, 10)
    self.assertEqual(result.my_model.foo, 7)

  def test_flag_config_dataclass_string_dict(self):
    result = _test_flags(_CONFIG, '.my_model.baz["foo.b"]=rab')
    self.assertEqual(result.my_model.baz['foo.b'], 'rab')

  def test_flag_config_dataclass_tuple_dict(self):
    result = _test_flags(_CONFIG, '.my_model.buz[(0,1)]=hello')
    self.assertEqual(result.my_model.buz[(0, 1)], 'hello')

  def test_flag_config_dataclass_typed_tuple(self):
    result = _test_flags(_CONFIG, '.my_model.boj=(0, 1)')
    self.assertEqual(result.my_model.boj, (0, 1))


class DataClassParseFnTest(absltest.TestCase):

  def test_parse_no_custom_value(self):
    result = _test_flags(
        _CONFIG, '.baseline_model.foo=10', parse_fn=parse_config_flag
    )
    self.assertEqual(result.my_model.foo, 3)
    self.assertEqual(result.baseline_model.foo, 10)

  def test_parse_custom_value_applied(self):
    result = _test_flags(
        _CONFIG, '=75', '.baseline_model.foo=10', parse_fn=parse_config_flag
    )
    self.assertEqual(result.my_model.foo, 75)
    self.assertEqual(result.baseline_model.foo, 10)

  def test_parse_custom_value_applied_no_explicit_parse_fn(self):
    result = _test_flags(CustomParserConfig(0), '=75', '.i=12')
    self.assertEqual(result.i, 12)
    self.assertEqual(result.j, 76)

  def test_parse_out_of_order(self):
    with self.assertRaises(config_flag_lib.FlagOrderError):
      _ = _test_flags(
          _CONFIG, '.baseline_model.foo=10', '=75', parse_fn=parse_config_flag
      )
    # Note: If this is ever supported, add verification that overrides are
    # applied correctly.

  def test_parse_assign_dataclass(self):
    flag_values = flags.FlagValues()

    def always_fail(v):
      raise ValueError()

    result = config_flags.DEFINE_config_dataclass(
        'test_config', _CONFIG, flag_values=flag_values, parse_fn=always_fail)
    flag_values(['program'])
    flag_values['test_config'].value = parse_config_flag('12')
    self.assertEqual(result.value.my_model.foo, 12)

  def test_parse_invalid_custom_value(self):
    with self.assertRaises(flags.IllegalFlagValueError):
      _ = _test_flags(
          _CONFIG, '=?', '.baseline_model.foo=10', parse_fn=parse_config_flag
      )

  def test_parse_overrides_applied(self):
    result = _test_flags(
        _CONFIG, '=34', '.my_model.foo=10', parse_fn=parse_config_flag
    )
    self.assertEqual(result.my_model.foo, 10)

if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Example of basic DEFINE_config_dict usage.

To run this example:
python define_config_dict_basic.py -- --my_config_dict.field1=8 \
  --my_config_dict.nested.field=2.1 --my_config_dict.tuple='(1, 2, (1, 2))'
"""

from absl import app

from ml_collections import config_dict
from ml_collections import config_flags

config = config_dict.ConfigDict()
config.field1 = 1
config.field2 = 'tom'
config.nested = config_dict.ConfigDict()
config.nested.field = 2.23
config.tuple = (1, 2, 3)

_CONFIG = config_flags.DEFINE_config_dict('my_config_dict', config)


def main(_):
  print(_CONFIG.value)


if __name__ == '__main__':
  app.run(main)
